from typing import Optional

import storch
from storch.method.method import Method

from storch.sampling import SampleWithoutReplacement


class ScoreFunctionWOR(Method):
    """
    Implement Buy 4 REINFORCE Samples, Get a Baseline for Free! https://openreview.net/pdf?id=r1lgTGL5DE
    Use biased=True for the biased normalized version which has lower variance.
    """

    def __init__(
        self, plate_name: str, k: int, biased: bool = True, use_baseline: bool = True
    ):
        # Use k + 1 to be able to compute kappa, the k+1th perturbed log-prob
        super().__init__(plate_name, SampleWithoutReplacement(plate_name, k, biased))
        self.biased = biased
        self.use_baseline = use_baseline

    def is_pathwise(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> bool:
        # We only want to add a loss on the stochastic tensor with the same plate as the cost node.
        # This is because the estimator computes the gradient with the respect to the JOINT log probability.
        # If we would have added the gradient for all stochastic tensors, these would just be duplicates of the same
        # loss being added (ie that gradient would be oversampled)
        # TODO: This should be rewritten using stochastic node partitions
        for distr_plate in tensor.plates:
            if distr_plate.name == self.plate_name:
                for cost_plate in cost_node.plates:
                    if cost_plate.name == self.plate_name:
                        if cost_plate is distr_plate:
                            return False
                        # TODO: What exactly does this mean?
                        return True
                raise ValueError(
                    "The given tensor contains an ancestral plate that the cost node doesn't have."
                )
        return True

    def multiplicative_estimator(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> Optional[storch.Tensor]:
        cost_plate = None
        for _p in cost_node.plates:
            if _p.name == self.plate_name:
                cost_plate = _p
                break
        if self.use_baseline:
            iw = self.sampling_method.compute_iw(cost_plate, biased=False)
            BS = storch.sum(iw * cost_node, cost_plate)
            probs = cost_plate.log_probs.exp()
            if self.biased:
                # Equation 11
                WS = storch.sum(iw, cost_plate)
                WiS = (WS - iw + probs).detach()
                diff_cost = cost_node - BS / WS
                return storch.sum(iw / WiS * diff_cost.detach(), cost_plate)
            else:
                # Equation 10
                weighted_cost = cost_node * (1 - probs + iw)
                diff_cost = weighted_cost - BS
                return storch.sum(iw * diff_cost.detach(), cost_plate)
        else:
            # Equation 9
            # TODO: This seems inefficient... The plate should already contain the IW, right? Same for above if not self.biased
            iw = self.sampling_method.compute_iw(cost_plate, self.biased)
            return storch.sum(cost_node.detach() * iw, self.plate_name)



import storch
import torch
from torch.distributions import Bernoulli, OneHotCategorical

expect = storch.method.Expect("x")
probs = torch.tensor([0.95, 0.01, 0.01, 0.01, 0.01, 0.01], requires_grad=True)
indices = torch.tensor([1.0, 2.0, 3.0, 4.0, 5.0, 6.0])
b = OneHotCategorical(probs=probs)
z = expect.sample(b)
c = (2.4 * z * indices).sum(-1)
storch.add_cost(c, "no_baseline_cost")

storch.backward()

expect_grad = z.grad["probs"].clone()


def eval(grads):
    print("----------------------------------")
    grad_samples = storch.gather_samples(grads, "variance")
    mean = storch.reduce_plates(grad_samples, plates=["variance"])
    print("mean grad", mean)
    print("expected grad", expect_grad)
    print("specific_diffs", (mean - expect_grad) ** 2)
    mse = storch.reduce_plates((grad_samples - expect_grad) ** 2).sum()
    print("MSE", mse)
    bias = (storch.reduce_plates((mean - expect_grad) ** 2)).sum()
    print("bias", bias)
    return bias


# method = storch.method.UnorderedSetEstimator("x", k=6)
# method = storch.REBAR()
# method = storch.method.ScoreFunction(
#     "x", baseline_factory="batch_average", n_samples=10
# )
# method = storch.method.ScoreFunction("x", baseline_factory="moving_average")
method = storch.method.RELAX("x", in_dim=6)
# method = storch.method.RELAX("x", in_dim=6)
optim = torch.optim.SGD(method.parameters(), lr=0.000001)
tot_bias = 0
for _ in range(100):
    grads = []
    print(method.eta, method.temperature)
    for i in range(100):
        optim.zero_grad()
        b = OneHotCategorical(probs=probs)
        z = method.sample(b)
        c = (2.4 * z * indices).sum(-1) + 100
        storch.add_cost(c, "baseline_cost")

        storch.backward(update_estimator_params=True)
        grad = z.grad["probs"].clone()
        grads.append(grad)
        if (i + 1) % 1000 == 0:
            print(i)
            eval(grads)
        optim.step()

    # print("FINAL EVAL")
    tot_bias += eval(grads)

print(tot_bias / 100)
# expected_estim_grad = 0.0
# for i in range(6):
#     to_predict = [0] * 6
#     to_predict[i] = 1
#     to_predict = torch.tensor(to_predict)
#     b = OneHotCategorical(probs=probs)
#     while True:
#         z = method.sample(b)
#         _z = z._tensor
#         if _z.ndim == 2:
#             _z = _z[0]
#         if torch.argmax(_z) == i:
#             break
#     c = (2.4 * z * indices).sum(-1) + 100
#     storch.add_cost(c, "baseline_cost")
#
#     storch.backward(update_estimator_params=False)
#     grad = z.grad["probs"].clone()
#     expected_estim_grad += grad * probs[i]
# print("estimated grad", expected_estim_grad)
# print("expected grad", expect_grad)
# bias = ((expected_estim_grad - expect_grad) ** 2).sum()
# print("bias", bias)



from typing import Optional

from storch.sampling import SamplingMethod
from torch.distributions import Distribution
import storch
import torch
from storch import Plate
import itertools


class Enumerate(SamplingMethod):
    def __init__(self, plate_name: str, budget=10000):
        super().__init__(plate_name)
        self.budget = budget

    def sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        requires_grad: bool,
    ) -> (storch.StochasticTensor, Plate):
        # TODO: Currently very inefficient as it isn't batched
        # TODO: What if the expectation has a parent
        if not distr.has_enumerate_support:
            raise ValueError(
                "Can only calculate the expected value for distributions with enumerable support."
            )
        support: torch.Tensor = distr.enumerate_support(expand=True)
        support_non_expanded: torch.Tensor = distr.enumerate_support(expand=False)
        expect_size = support.shape[0]

        batch_len = len(plates)
        sizes = support.shape[
            batch_len + 1 : len(support.shape) - len(distr.event_shape)
        ]
        amt_samples_used = expect_size
        cross_products = 1 if not sizes else None
        for dim in sizes:
            amt_samples_used = amt_samples_used ** dim
            if not cross_products:
                cross_products = dim
            else:
                cross_products = cross_products ** dim

        if amt_samples_used > self.budget:
            raise ValueError(
                "Computing the expectation on this distribution would exceed the computation budget."
            )

        enumerate_tensor = support.new_zeros(
            [amt_samples_used] + list(support.shape[1:])
        )
        support_non_expanded = support_non_expanded.squeeze().unsqueeze(1)
        for i, t in enumerate(
            itertools.product(support_non_expanded, repeat=cross_products)
        ):
            enumerate_tensor[i] = torch.cat(t, dim=0)

        enumerate_tensor = enumerate_tensor.detach()

        plate_size = enumerate_tensor.shape[0]

        plate = Plate(self.plate_name, plate_size, plates.copy())
        plates.insert(0, plate)

        s_tensor = storch.StochasticTensor(
            enumerate_tensor,
            parents,
            plates,
            self.plate_name,
            plate_size,
            distr,
            requires_grad,
        )
        return s_tensor, plate

    def weighting_function(
        self, tensor: storch.StochasticTensor, plate: Plate
    ) -> Optional[storch.Tensor]:
        # Weight by the probability of each possible event
        log_probs = tensor.distribution.log_prob(tensor)
        if log_probs.plate_dims < len(log_probs.shape):
            log_probs = log_probs.sum(
                dim=list(range(tensor.plate_dims, len(log_probs.shape)))
            )
        return log_probs.exp()



"""
Based on Yin and Zhou, ICLR 2019
"""

import torch
import storch
from torch.distributions import Bernoulli
from storch.method import ScoreFunction, Expect, UnorderedSetEstimator

p = torch.tensor([0.49, 0.499, 0.501, 0.51])
eta = torch.tensor([0.0, 0.0, 0.0, 0.0], requires_grad=True)
optim = torch.optim.SGD([eta], lr=1.0)


def experiment(method):
    for i in range(2000):
        optim.zero_grad()
        b = Bernoulli(logits=eta)
        x = method(b)
        cost = torch.sum((x - p) ** 2, -1)
        storch.add_cost(cost, "cost")
        storch.backward()
        optim.step()
        if i % 100 == 0:
            print(eta)
        # return eta.grad.clone()


experiment(Expect("x"))
# experiment(UnorderedSetEstimator("x", 5))
# print("True gradient", true_grad)

# for n in range(2, 9):
#     # _method = ScoreFunction("x", n, baseline_factory="batch_average")
#     _method = UnorderedSetEstimator("x", n)
#     # gradients = []
#     for i in range(1000):
#         experiment(_method)
#     # gradients = torch.cat(gradients)
#     # print(n)
#     # print("variance", gradients.var(0))
#     # print("bias", torch.sum((gradients.mean(0) - true_grad) ** 2))



from __future__ import annotations
from typing import Optional, Union, List, Callable, Tuple

import storch
import torch
from torch.distributions import Distribution, Gumbel
import itertools
from storch.sampling.method import SamplingMethod
from storch.sampling.seq import IterDecoding, AncestralPlate, right_expand_as


class SampleWithoutReplacement(IterDecoding):
    """
    Sampling method for sampling without replacement from (sequences of) discrete distributions.
    Implements Stochastic Beam Search https://arxiv.org/abs/1903.06059 with the weighting as defined by
    REINFORCE without replacement https://openreview.net/forum?id=r1lgTGL5DE
    """

    EPS = 1e-8

    def __init__(self, plate_name: str, k: int, biased_iw: bool = False, eos=None):
        super().__init__(plate_name, k, eos)
        if k < 2:
            raise ValueError(
                "Can only sample with replacement for more than 1 samples."
            )
        self.biased_iw = biased_iw

    def reset(self):
        super().reset()
        # Cumulative perturbed log probabilities of the samples
        self.perturbed_log_probs: Optional[storch.Tensor] = None

    def decode_step(
        self,
        indices: Tuple[int],
        yv_log_probs: storch.Tensor,
        joint_log_probs: Optional[storch.Tensor],
        sampled_support_indices: Optional[storch.Tensor],
        parent_indexing: Optional[storch.Tensor],
        is_conditional_sample: bool,
        amt_plates: int,
        amt_samples: int,
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor):
        """
        Decode given the input arguments for a specific event using stochastic beam search.
        :param indices: Tuple of integers indexing the current event to sample.
        :param yv_log_probs:  Log probabilities of the different options for this event. distr_plates x k? x |D_yv|
        :param joint_log_probs: The log probabilities of the samples so far. None if `not is_conditional_sample`. prev_plates x amt_samples
        :param sampled_support_indices: Tensor of samples so far. None if this is the first set of indices. plates x k x events
        :param parent_indexing: Tensor indexing the parent sample. None if `not is_conditional_sample`.
        :param is_conditional_sample: True if a parent has already been sampled. This means the plates are more complex!
        :param amt_plates: The total amount of plates in both the distribution and the previously sampled variables
        :param amt_samples: The amount of active samples.
        :return: 3-tuple of `storch.Tensor`. 1: sampled_support_indices, with `:, indices` referring to the indices for the support.
        2: The updated `joint_log_probs` of the samples.
        3: The updated `parent_indexing`. How the samples index the parent samples. Can just return parent_indexing if nothing happens.
        4: The amount of active samples after this step.
        """

        first_sample = False
        if joint_log_probs is None:
            # We also know that k? is not present, so distr_plates x |D_yv|
            all_joint_log_probs = yv_log_probs
            # First condition on max being 0:
            self.perturbed_log_probs = 0.0
            first_sample = True
        elif is_conditional_sample > 0:
            # Make sure we are selecting the correct log-probabilities. As parents have been selected, this might change!
            # plates x amt_samples x |D_yv|
            yv_log_probs = yv_log_probs.gather(
                dim=-2,
                index=right_expand_as(
                    # Use the parent_indexing to select the correct plate samples. Make sure to limit to amt_samples!
                    parent_indexing[..., :amt_samples],
                    yv_log_probs,
                ),
            )
            # self.joint_log_probs: prev_plates x amt_samples
            # plates x amt_samples x |D_yv|
            all_joint_log_probs = joint_log_probs.unsqueeze(-1) + yv_log_probs
        else:
            # self.joint_log_probs: plates x amt_samples
            # plates x amt_samples x |D_yv|
            all_joint_log_probs = joint_log_probs.unsqueeze(
                -1
            ) + yv_log_probs.unsqueeze(-2)

        # Sample plates x k? x |D_yv| conditional Gumbel variables
        cond_G_yv = cond_gumbel_sample(all_joint_log_probs, self.perturbed_log_probs)

        # If there are finished samples, ensure eos is always sampled.
        if self.finished_samples is not None:
            # TODO: Is this the correct way of ensuring self.eos is always sampled for finished sequences?
            #  Coudl it bias things in any way?
            # Set the probability of continuing on finished sequences to -infinity so that they are filtered out during topk.
            # amt_finished
            finished_perturb_log_probs = self.perturbed_log_probs._tensor[
                self.finished_samples._tensor
            ]
            # amt_finished x |D_yv|
            finished_vec = finished_perturb_log_probs.new_full(
                (finished_perturb_log_probs.shape[0], cond_G_yv.shape[-1],),
                -float("inf"),
            )
            # Then make sure the log probability of the eos token is equal to the last perturbed log prob.
            finished_vec[:, self.eos] = finished_perturb_log_probs
            cond_G_yv[self.finished_samples] = finished_vec

        if not first_sample:
            # plates x (k * |D_yv|) (k == prev_amt_samples, in this case)
            cond_G_yv = cond_G_yv.reshape(cond_G_yv.shape[:-2] + (-1,))

        # Select the samples given the perturbed log probabilities
        self.perturbed_log_probs, arg_top = self.select_samples(
            cond_G_yv, all_joint_log_probs
        )
        amt_samples = arg_top.shape[-1]

        if first_sample:
            # plates x amt_samples
            joint_log_probs = all_joint_log_probs.gather(dim=-1, index=arg_top)
            # Index for the selected samples. Uses slice(amt_samples) for the first index in case k > |D_yv|
            # (:) * amt_plates + (indices for events) + amt_samples
            indexing = (slice(None),) * amt_plates + (slice(0, amt_samples),) + indices
            sampled_support_indices[indexing] = arg_top
        else:
            # Gather corresponding joint log probabilities. First reshape like previous to plates x (k * |D_yv|).
            joint_log_probs = all_joint_log_probs.reshape(cond_G_yv.shape).gather(
                dim=-1, index=arg_top
            )

            # |D_yv|
            size_domain = yv_log_probs.shape[-1]

            # Keep track of what parents were sampled for the arg top
            # plates x amt_samples
            chosen_parents = arg_top // size_domain
            sampled_support_indices = sampled_support_indices.gather(
                dim=amt_plates,
                index=right_expand_as(chosen_parents, sampled_support_indices),
            )
            if parent_indexing is not None:
                parent_indexing = parent_indexing.gather(dim=-1, index=chosen_parents)
            # Index for the selected samples. Uses slice(amt_samples) for the first index in case k > |D_yv|
            # plates x amt_samples
            chosen_samples = arg_top.remainder(size_domain)
            indexing = (slice(None),) * amt_plates + (slice(0, amt_samples),) + indices
            sampled_support_indices[indexing] = chosen_samples
        return sampled_support_indices, joint_log_probs, parent_indexing, amt_samples

    def select_samples(
        self, perturbed_log_probs: storch.Tensor, joint_log_probs: storch.Tensor,
    ) -> (storch.Tensor, storch.Tensor):
        """
        Given the perturbed log probabilities and the joint log probabilities of the new options, select which one to
        use for the sample.
        :param perturbed_log_probs: plates x (k? * |D_yv|). Perturbed log-probabilities. k is present if first_sample.
        :param joint_log_probs: plates x k? x (k? * |D_yv|). Joint log probabilities of the options. k is present if first_sample.
        :param first_sample:
        :return:
        """

        # We can sample at most the amount of what we previous sampled, combined with every option in the current domain
        # That is: prev_amt_samples * |D_yv|.
        amt_samples = min(self.k, perturbed_log_probs.shape[-1])
        # Take the top k over conditional perturbed log probs
        # plates x amt_samples
        return torch.topk(perturbed_log_probs, amt_samples, dim=-1)

    def create_plate(self, plate_size: int, plates: [storch.Plate]) -> AncestralPlate:
        plate = super().create_plate(plate_size, plates)
        plate.perturb_log_probs = storch.Tensor(
            self.perturbed_log_probs._tensor,
            [self.perturbed_log_probs],
            self.perturbed_log_probs.plates + [plate],
        )
        return plate

    def weighting_function(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        return self.compute_iw(plate, self.biased_iw).detach()

    def compute_iw(self, plate: AncestralPlate, biased: bool):
        # Compute importance weights. The kth sample has 0 weight, and is only used to compute the importance weights
        q = (
            1
            - (
                -(
                    plate.log_probs
                    - plate.perturb_log_probs._tensor[..., self.k - 1].unsqueeze(-1)
                ).exp()
            ).exp()
        ).detach()
        iw = plate.log_probs.exp() / (q + self.EPS)
        # Set the weight of the kth sample (kappa) to 0.
        iw[..., self.k - 1] = 0.0
        if biased:
            WS = storch.sum(iw, plate).detach()
            return iw / WS
        return iw

    def on_plate_already_present(self, plate: storch.Plate):
        if (
            not isinstance(plate, AncestralPlate)
            or plate.variable_index > self.variable_index
            or plate.n > self.k
        ):
            super().on_plate_already_present(plate)

    def set_mc_sample(
        self,
        new_sample_func: Callable[
            [Distribution, [storch.Tensor], [storch.Plate], int], torch.Tensor
        ],
    ) -> SamplingMethod:
        raise RuntimeError(
            "Cannot set monte carlo sampling for sampling without replacement."
        )


def log1mexp(a: torch.Tensor) -> torch.Tensor:
    """See appendix A of http://jmlr.org/papers/v21/19-985.html.
    Numerically stable implementation of log(1-exp(a))"""
    c = -0.693
    a1 = -a.abs()
    return torch.where(a1 > c, torch.log(-a1.expm1()), torch.log1p(-a1.exp()))


@storch.deterministic
def cond_gumbel_sample(all_joint_log_probs, perturbed_log_probs) -> torch.Tensor:
    # Sample plates x k? x |D_yv| Gumbel variables
    gumbel_d = Gumbel(loc=all_joint_log_probs, scale=1.0)
    G_yv = gumbel_d.rsample()

    # Condition the Gumbel samples on the maximum of previous samples
    # plates x k
    Z = G_yv.max(dim=-1)[0]
    T = perturbed_log_probs
    vi = T - G_yv + log1mexp(G_yv - Z.unsqueeze(-1))
    # plates (x k) x |D_yv|
    return T - vi.relu() - torch.nn.Softplus()(-vi.abs())


class SumAndSample(SampleWithoutReplacement):
    """
    Sums over S probable samples according to beam search and K sampled values that are not in the probable samples,
    then normalizes them accordingly.
    """

    def __init__(
        self,
        plate_name: str,
        sum_size: int,
        sample_size: int = 1,
        without_replacement: bool = False,
        eos=None,
    ):
        super().__init__(plate_name, sum_size + sample_size, eos=eos)
        self.sum_size = sum_size
        self.sample_size = sample_size
        if sum_size < 1 or sample_size < 1:
            raise ValueError("sum_size and sample_size should both be at least 1.")

    def select_samples(
        self, perturbed_log_probs: storch.Tensor, joint_log_probs: storch.Tensor,
    ) -> (storch.Tensor, storch.Tensor):
        # Select sum_size samples using joint log probs, and sample_size samples using perturbed joint log probs.

        # We can sample at most the amount of what we previous sampled, combined with every option in the current domain
        # That is: prev_amt_samples * |D_yv|.
        amt_sum = min(self.sum_size, joint_log_probs.shape[-1])
        # Take the top sum_size over the joint log probs. This is like beam search
        # plates x amt_samples
        _, sum_samples = torch.topk(joint_log_probs, amt_sum, dim=-1)
        sum_perturbed_log_probs = perturbed_log_probs[sum_samples]
        if amt_sum < self.sum_size:
            return sum_perturbed_log_probs, sum_samples

        # Not sure if this is the most efficient implementation
        # Should be positive, by the previous conditional.
        amt_sample = min(
            self.sample_size, perturbed_log_probs.shape[-1] - self.sum_size
        )
        sample_perturbed_log_probs, samples = torch.topk(
            joint_log_probs, amt_sample + perturbed_log_probs.shape[-1], dim=-1
        )

        # TODO: This isn't finished yet.



import storch
import torch
from torch.distributions import Bernoulli
from storch.method import GumbelSoftmax

torch.manual_seed(0)

p = torch.tensor(0.5, requires_grad=True)

for i in range(900):
    sample = GumbelSoftmax(f"sample_{i}")(Bernoulli(p))
    storch.add_cost(sample, f"cost_{i}")

storch.backward()


from storch.tensor import Plate
import pytest
import torch
import storch

n = {1, 5, 1000, -1}


@pytest.fixture
def plate():
    return Plate("test", 123, [Plate("parent", 31, [])])


def test_plate(plate):
    assert False


def test_reduce(plate):
    assert False


def test_on_collecting_args(plate):
    assert False


@pytest.mark.parametrize("n", n)
def test_on_unwrap_tensor(plate, n):
    assert False


def to_storch(tensor: torch.Tensor) -> storch.Tensor:
    return storch.Tensor(tensor, [], [], "test")


@pytest.fixture
def a() -> torch.Tensor:
    return torch.tensor([0.1, 0.2, 0.3, 0.4])


@pytest.fixture
def b() -> torch.Tensor:
    return torch.tensor([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])


@pytest.fixture
def mask(request):
    return torch.tensor(request.param)


@pytest.fixture
def expected(request):
    return torch.tensor(request.param)


@pytest.fixture
def value(request):
    return torch.tensor(request.param)


def c1() -> torch.Tensor:
    return torch.tensor([0.6, 0.6, 0.6])


def test_bool(a):
    a = to_storch(a)
    with pytest.raises(storch.IllegalStorchExposeError):
        if a[0] < 0.3:
            assert False


@pytest.mark.parametrize(
    "mask,expected",
    [
        ([False, False, True, False], 0.3),
        (2, 0.3),
        ([1, 3], [0.2, 0.4]),
        ([True, True, False, True], [0.1, 0.2, 0.4]),
    ],
    indirect=True,
)
def test_get_item(a, mask, expected):
    a = to_storch(a)
    assert (a[mask]._tensor == expected).all()


@pytest.mark.parametrize(
    "mask,value,expected",
    [
        ([False, False, True, False], [0.6], [0.1, 0.2, 0.6, 0.4]),
        ([True, True, False, True], [0.6, 0.6, 0.6], [0.6, 0.6, 0.3, 0.6]),
        (2, [0.6], [0.1, 0.2, 0.6, 0.4]),
        ([0, 1, 3], [0.6, 0.6, 0.6], [0.6, 0.6, 0.3, 0.6]),
    ],
    indirect=True,
)
def test_set_item(a, mask, value, expected):
    a = to_storch(a)
    a[mask] = value
    assert (a._tensor == expected).all()


@pytest.mark.parametrize(
    "mask_b,value,expected",
    [
        ((1, 2), 0.7, [[0.1, 0.2, 0.3], [0.4, 0.5, 0.7]]),
        (
            torch.tensor([[True, False, True], [False, True, False]]),
            0.7,
            [[0.7, 0.2, 0.7], [0.4, 0.7, 0.6]],
        ),
    ],
    indirect=["value", "expected"],
)
def test_set_item_2(b, mask_b, value, expected):
    b = to_storch(b)
    b[mask_b] = value
    assert (b._tensor == expected).all()



from typing import List, Union, Optional, Tuple

import storch
from torch.distributions.utils import clamp_probs
import torch

from storch.typing import AnyTensor, _indices, _plates


def _convert_indices(tensor: storch.Tensor, dims: _indices) -> (List[int], List[str]):
    conv_indices = []
    red_batches = []
    if not isinstance(dims, List):
        dims = [dims]
    for index in dims:
        if isinstance(index, int):
            if index >= tensor.plate_dims or index < 0 and index >= -tensor.event_dims:
                conv_indices.append(index)
            else:
                print(tensor.shape, index)
                raise IndexError(
                    "Can only pass indexes for event dimensions."
                    + str(tensor)
                    + ". Index: "
                    + str(index)
                )
        else:
            if isinstance(index, storch.Plate):
                index = index.name
            conv_indices.append(tensor.get_plate_dim_index(index))
            red_batches.append(index)
    return tuple(conv_indices), red_batches


def mean(tensor: storch.Tensor, dims: _indices) -> storch.Tensor:
    indices, reduced_batches = _convert_indices(tensor, dims)
    return storch.reduce(torch.mean, plates=reduced_batches)(tensor, indices)


def sum(tensor: storch.Tensor, dims: _indices) -> storch.Tensor:
    indices, reduced_batches = _convert_indices(tensor, dims)
    return storch.reduce(torch.sum, plates=reduced_batches)(tensor, indices)


def logsumexp(tensor: storch.Tensor, dims: _indices) -> storch.Tensor:
    indices, reduced_batches = _convert_indices(tensor, dims)
    return storch.reduce(torch.logsumexp, plates=reduced_batches)(tensor, indices)


def expand_as(tensor: AnyTensor, expand_as: AnyTensor) -> AnyTensor:
    return storch.deterministic(torch.expand_as)(tensor, expand_as)


def _handle_inputs(
    tensor: AnyTensor, plates: Optional[_plates],
) -> (storch.Tensor, List[storch.Plate]):
    if isinstance(plates, storch.Plate):
        plates = [plates]
    if not isinstance(tensor, storch.Tensor):
        if not plates:
            raise ValueError("Make sure to pass plates when passing a torch.Tensor.")
        index_tensor = 0

        for plate in plates:
            if not isinstance(plate, storch.Plate):
                raise ValueError(
                    "Cannot handle plate names when passing a torch.Tensor"
                )
            if plate.n > 1:
                if tensor.shape[index_tensor] != plate.n:
                    raise ValueError(
                        "Received a tensor that does not align with the given plates."
                    )
                index_tensor += 1
        return storch.Tensor(tensor, [], plates), plates

    if not plates:
        return tensor, tensor.plates
    if isinstance(plates, str):
        return tensor, [tensor.get_plate(plates)]
    r_plates = []
    for plate in plates:
        if isinstance(plate, storch.Plate):
            r_plates.append(plate)
        else:
            r_plates.append(tensor.get_plate(plate))
    return tensor, r_plates


def gather(input: storch.Tensor, dim: str, index: storch.Tensor):
    # TODO: Should be allowed to accept int and storch.Plate as well
    return storch.deterministic(torch.gather, dim=dim, expand_plates=True)(
        input, index=index
    )


def reduce_plates(
    tensor: AnyTensor, plates: Optional[_plates] = None, detach_weights=True,
) -> storch.Tensor:
    """
    Reduce the tensor along the given plates. This takes into account how different samples are weighted, and should
    nearly always be used instead of reducing plate dimensions using the mean or the sum.
    By default, this reduces all plates.

    Args:
        tensor: Tensor to reduce
        plates: Plates to reduce. If None, this reduces all plates (default). Can be a string, Plate, or list of string
        and Plates.
        detach_weights: Whether to detach the weighting of the samples from the graph

    Returns:
        The reduced tensor
    """
    tensor, plates = _handle_inputs(tensor, plates)
    for plate in order_plates(plates, reverse=True):
        if plate.n > 1:
            tensor = plate.reduce(tensor, detach_weights=detach_weights)
    return tensor


def _isroot(plate: storch.Plate, plates: [storch.Plate]):
    """
    Returns true if the plate is a root of the list of plates.
    """
    for parent in plate.parents:
        if parent in plates:
            return False
        if not _isroot(parent, plates):
            return False
    return True


def order_plates(plates: [storch.Plate], reverse=False):
    """
    Topologically order the given plates.
    Uses Kahn's algorithm.
    """
    sorted = []
    roots = []
    in_edges = {}
    out_edges = {p.name: [] for p in plates}
    for p in plates:
        if _isroot(p, plates):
            roots.append(p)
        in_edges[p.name] = p.parents.copy()
        for _p in p.parents:
            if _p.name in out_edges:
                out_edges[_p.name].append(p)
            # This is possible if the input list of plates does not contain the parent. We still need to register it
            # to make sure the list is sorted in global topological ordering!
            else:
                out_edges[_p.name] = []
    while roots:
        n = roots.pop()
        if n in plates:
            sorted.append(n)
        for m in out_edges[n.name]:
            remaining_edges = in_edges[m.name]
            remaining_edges.remove(n)
            if not remaining_edges:
                roots.append(m)
    for remaining_edges in in_edges.values():
        if remaining_edges and any(map(lambda p: p in plates, remaining_edges)):
            raise ValueError("List of plates contains a cycle")
    if reverse:
        return reversed(sorted)
    return sorted


def variance(
    tensor: AnyTensor,
    variance_plate: Union[storch.Plate, str],
    plates: Optional[_plates] = None,
    detach_weights=True,
) -> storch.Tensor:
    """
    Compute the variance of the tensor along the plate dimensions. This takes into account how different samples are weighted.

    Args:
        tensor: Tensor to compute variance over
        plates: Plates to reduce.
        detach_weights: Whether to detach the weighting of the samples from the graph
    Returns:
        The variance of the tensor.
    """
    tensor, plates = _handle_inputs(tensor, plates)
    found_plate = False
    for plate in plates:
        if plate == variance_plate:
            found_plate = True
            break
        if plate.name == variance_plate:
            variance_plate = plate
            found_plate = True
            break
    if not found_plate:
        raise ValueError(
            "Should pass a variance_plate that is included in the passed plates."
        )
    mean = variance_plate.reduce(tensor, detach_weights=detach_weights)
    variance = (tensor - mean) ** 2

    event_dims = tensor.event_dim_indices
    if len(event_dims) > 0:
        variance = variance.sum(tensor.event_dim_indices)
    return reduce_plates(variance, detach_weights=detach_weights)


def grad(
    outputs,
    inputs,
    grad_outputs=None,
    retain_graph: Optional[bool] = None,
    create_graph: bool = False,
    only_inputs: bool = True,
    allow_unused: bool = False,
) -> Tuple[storch.Tensor, ...]:
    """
    Helper method for computing torch.autograd.grad on storch tensors. Returns storch Tensors as well.
    """
    args, _, _, _ = storch.wrappers._prepare_args(
        [outputs, inputs, grad_outputs], {}, unwrap=True, align_tensors=False,
    )
    _outputs, _inputs, _grad_outputs = tuple(args)
    grads = torch.autograd.grad(
        _outputs,
        _inputs,
        grad_outputs=_grad_outputs,
        retain_graph=retain_graph,
        create_graph=create_graph,
        only_inputs=only_inputs,
        allow_unused=allow_unused,
    )
    storch_grad = []
    for i, grad in enumerate(grads):
        input = inputs[i]
        if isinstance(input, storch.Tensor):
            storch_grad.append(
                storch.Tensor(
                    grad, outputs + [input], input.plates, input.name + "_grad"
                )
            )
        else:
            storch_grad.append(storch.Tensor(grad, outputs, [], "grad"))
    return tuple(storch_grad)


@storch.deterministic(expand_plates=True)
def cat(*args, **kwargs):
    """
    Version of :func:`torch.cat` that is compatible with :class:`storch.Tensor`.
    Required because :meth:`torch.Tensor.__torch_function__` is not properly implemented for :func:`torch.cat`:
    https://github.com/pytorch/pytorch/issues/34294

    """
    return torch.cat(*args, **kwargs)


@storch.deterministic
def conditional_gumbel_rsample(
    hard_sample: torch.Tensor, probs: torch.Tensor, bernoulli: bool, temperature,
) -> torch.Tensor:
    """
    Conditionally re-samples from the distribution given the hard sample.
    This samples z \sim p(z|b), where b is the hard sample and p(z) is a gumbel distribution.
    """
    # Adapted from torch.distributions.relaxed_bernoulli and torch.distributions.relaxed_categorical
    shape = hard_sample.shape

    probs = clamp_probs(probs.expand_as(hard_sample))
    v = clamp_probs(torch.rand(shape, dtype=probs.dtype, device=probs.device))
    if bernoulli:
        pos_probs = probs[hard_sample == 1]
        v_prime = torch.zeros_like(hard_sample)
        # See https://arxiv.org/abs/1711.00123
        v_prime[hard_sample == 1] = v[hard_sample == 1] * pos_probs + (1 - pos_probs)
        v_prime[hard_sample == 0] = v[hard_sample == 0] * (1 - probs[hard_sample == 0])
        log_sample = (
            probs.log() + probs.log1p() + v_prime.log() + v_prime.log1p()
        ) / temperature
        return log_sample.sigmoid()
    # b=argmax(hard_sample)
    b = hard_sample.max(-1).indices
    # b = F.one_hot(b, hard_sample.shape[-1])

    # See https://arxiv.org/abs/1711.00123
    log_v = v.log()
    # i != b (indexing could maybe be improved here, but i doubt it'd be more efficient)
    log_v_b = torch.gather(log_v, -1, b.unsqueeze(-1))
    cond_gumbels = -(-torch.div(log_v, probs) - log_v_b).log()
    # i = b
    index_sample = hard_sample.bool()
    cond_gumbels[index_sample] = -(-log_v[index_sample]).log()
    scores = cond_gumbels / temperature
    return (scores - scores.logsumexp(dim=-1, keepdim=True)).exp()



from functools import reduce
from operator import mul
from typing import Optional, Callable, Tuple

import torch
from torch.distributions import Distribution, Bernoulli

from torch.nn import Parameter

import storch
from storch import (
    Plate,
    CostTensor,
    StochasticTensor,
    deterministic,
    conditional_gumbel_rsample,
)
from storch.sampling import MonteCarlo, SamplingMethod
from storch.typing import Dims
from storch.method.method import Reparameterization, Method

import torch.nn.functional as F

from storch.util import get_distr_parameters, rsample_gumbel_softmax, split, magic_box


class Baseline(torch.nn.Module):
    def __init__(self, in_dim: Dims):
        super().__init__()
        self.reshape = False
        if not isinstance(in_dim, int):
            self.reshape = True
            in_dim = reduce(mul, in_dim)
        self.in_dim = in_dim
        self.fc1 = torch.nn.Linear(in_dim, 50)
        self.fc2 = torch.nn.Linear(50, 1)

    def forward(self, x: storch.Tensor):
        if self.reshape or self.in_dim <= 1:
            x = x.reshape(x.shape[: x.plate_dims] + (-1,))
        return self.fc2(F.relu(self.fc1(x))).squeeze(-1)


class LAX(Reparameterization):
    """
    Gradient estimator for continuous random variables.
    Implements the LAX estimator from Grathwohl et al, 2018 https://arxiv.org/abs/1711.00123
    Code inspired by https://github.com/duvenaud/relax/blob/master/pytorch_toy.py
    """

    def __init__(
        self,
        plate_name: str,
        *,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
        c_phi: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        in_dim=None,
    ):
        """
        Either c_phi or in_dim needs to be non-None!
        :param plate_name: Name of the plate
        :param sampling_method: The sampling method to use
        :param n_samples: The amount of samples to take
        :param c_phi: The baseline network. Needs to be specified if `in_dim`  is not specified.
        :param in_dim: The size of the default baseline. Needs to be set if `c_phi` is not specified.
        """
        super().__init__(plate_name, sampling_method, n_samples)
        if c_phi:
            self.c_phi = c_phi
        else:
            self.c_phi = Baseline(in_dim)
        # TODO: Add baseline strength

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        """
        We do a reparameterized sample, but we have to make sure we detach afterwards. The rsample is to make sure we
        can backpropagate through the sample in the estimator.
        """
        return tensor.detach()

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        # Input rsampled value into c_phi
        output_baseline = self.c_phi(tensor)

        # Compute log probability. Make sure not to use the rsampled value: We want to compute the distribution
        # not through the sample but only through the distributional parameters.
        log_prob = tensor.distribution.log_prob(tensor.detach())
        log_prob = log_prob.sum(dim=tensor.event_dim_indices)

        diff = cost_node - output_baseline  # [(...,) + (None,) * d_log_prob.event_dims]
        var_loss = 0.0
        # TODO: derivs is missing?
        for param, d_log_prob, d_output_baseline in derivs:
            # Compute total derivative with respect to the parameter
            d_param = diff * d_log_prob + d_output_baseline
            # Reduce the plate of this sample in case multiple samples are taken
            d_param = storch.reduce_plates(d_param, plates=[tensor.name])
            # Compute backwards from the parameters using its total derivative
            if isinstance(param, storch.Tensor):
                param = param._tensor
            param.backward(d_param._tensor, retain_graph=True)
            # Compute the gradient variance
            variance = (d_param ** 2).sum(d_param.event_dim_indices)
            var_loss += storch.reduce_plates(variance)

        c_phi_params = []

        for param in self.c_phi.parameters(recurse=True):
            if param.requires_grad:
                c_phi_params.append(param)
        d_variance = torch.autograd.grad([var_loss._tensor], c_phi_params)
        for i in range(len(c_phi_params)):
            c_phi_params[i].backward(d_variance[i])
        return log_prob

    def is_pathwise(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        return False

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor)]
    ) -> None:
        # During the normal backwards call, the parameters are accumulated gradients to the control variate parameters.
        # We don't want to minimize wrt to that loss, but to the one we define here.
        for param in self.control_params:
            param.grad = None
        tensors = []
        for tensor, _ in result_triples:
            if tensor not in tensors:
                tensors.append(tensor)
        # minimize the variance of the gradient with respect to the input parameters
        for tensor in tensors:
            d_param = next(iter(tensor.grad.values()))
            variance = (d_param ** 2).sum(d_param.event_dim_indices)
            var_loss = storch.reduce_plates(variance)

            d_variance = torch.autograd.grad(
                [var_loss._tensor], self.control_params, retain_graph=True,
            )
            print(d_variance)

            for i in range(len(self.control_params)):
                self.control_params[i].backward(d_variance[i])

    def should_create_higher_order_graph(self) -> bool:
        return True


def discretize(tensor: torch.Tensor, distr: Distribution) -> torch.Tensor:
    # Adapted from pyro.relaxed_straight_through
    if isinstance(distr, Bernoulli):
        return tensor.round()
    argmax = tensor.max(-1)[1]
    hard_sample = torch.zeros_like(tensor)
    if argmax.dim() < hard_sample.dim():
        argmax = argmax.unsqueeze(-1)
    return hard_sample.scatter_(-1, argmax, 1)


class RELAX(Method):
    """
    Gradient estimator for Bernoulli and Categorical distributions on any function.
    Implements the RELAX estimator from Grathwohl et al, 2018, https://arxiv.org/abs/1711.00123
    and the REBAR estimator from Tucker et al, 2017, https://arxiv.org/abs/1703.07370
    Code inspired by https://github.com/duvenaud/relax/blob/master/pytorch_toy.py
    """

    def __init__(
        self,
        plate_name: str,
        *,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
        c_phi: Callable[[torch.Tensor], torch.Tensor] = None,
        in_dim: Dims = None,
        rebar=False,
        train_eta=True,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(
            plate_name,
            sampling_method.set_mc_sample(self.sample_gumbel).set_mc_weighting_function(
                 self.plate_weighting
            ),
        )
        if c_phi:
            self.c_phi = c_phi
        elif in_dim:
            self.c_phi = Baseline(in_dim)
        else:
            raise ValueError("Either pass an explicit control variate c_phi or an input dimension")

        self.temperature = Parameter(torch.tensor(1.0))
        self.rebar = rebar
        self.eta = 1.0

        self.control_params = [self.temperature]
        if train_eta:
            self.eta = Parameter(torch.tensor(self.eta))
            self.control_params.append(self.eta)

        if isinstance(self.c_phi, torch.nn.Module):
            for c_phi_param in self.c_phi.parameters(recurse=True):
                if c_phi_param.requires_grad:
                    self.control_params.append(c_phi_param)

    def sample_gumbel(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ):
        if self.rebar:
            relaxed_sample = rsample_gumbel_softmax(
                distr, amt_samples, self.temperature, straight_through=False
            )
            # TODO: For rebar, what if sample from a sequence, and amt_samples is 1? Then three are sampled (for each one, also hard_sample and cond_sample.)
            #    That'd still blow up... But I guess it's required?
            # In REBAR, the objective function is evaluated for the Gumbel sample, the conditional Gumbel sample \tilde{z} and the argmax of the Gumbel sample.
            hard_sample = discretize(relaxed_sample, distr)
            cond_sample = conditional_gumbel_rsample(
                hard_sample, distr.probs, isinstance(distr, torch.distributions.Bernoulli), self.temperature
            )

            # return (H(z), z, z|H(z)
            return torch.cat([hard_sample, relaxed_sample, cond_sample], 0)
        return rsample_gumbel_softmax(
            distr, amt_samples, self.temperature, straight_through=False
        )

    def plate_weighting(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        if self.rebar:
            # Only weight over the true samples, put the relaxed samples to weight 0. This also makes sure
            # that they will not be backpropagated through in the cost backwards pass
            n = int(tensor.n // 3)
            weighting = tensor._tensor.new_zeros((tensor.n,))
            weighting[:n] = tensor._tensor.new_tensor(1.0 / n)
            return weighting
        return None

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        if self.rebar:
            # Make sure the r-sampled gumbels don't backpropagate in the normal backward pass
            # TODO: This doesn't work. We need to record the gradients with respect to the cost, as they are used in
            # the estimator. By detaching it, they are no longer recorded. However, not detaching means the normal loss
            # will capture it... Need to rethink this.
            return tensor
        else:
            # Return H(z) for the function evaluation if using RELAX
            return discretize(tensor, tensor.distribution)

    def compute_estimator(
        self, tensor: StochasticTensor, cost: CostTensor, plate_index: int, n: int, distribution: Distribution
    ) -> Tuple[storch.Tensor, storch.Tensor, storch.Tensor]:
        if self.rebar:
            log_prob = distribution.log_prob(tensor)
            @storch.deterministic
            def compute_REBAR_cvs(estimator: RELAX, tensor: StochasticTensor, cost: CostTensor):
                # TODO: Does it make more sense to implement REBAR by adding another plate dimension
                #  that controls the three different types of samples? (hard, relaxed, cond)?
                #  Then this can be implemented by simply reducing that plate, ie not requiring the weird zeros
                # Split the sampled tensor
                empty_slices = (slice(None),) * plate_index
                _index1 = empty_slices + (slice(n),)
                _index2 = empty_slices + (slice(n, 2 * n),)
                _index3 = empty_slices + (slice(2 * n, 3 * n),)
                relaxed_sample, cond_sample = (
                        tensor[_index2],
                        tensor[_index3],
                    )
                relaxed_cost, cond_cost = cost[_index2], cost[_index3]

                # Compute the control variates and log probabilities for the samples
                _c_phi_relaxed = estimator.c_phi(relaxed_sample) + relaxed_cost
                _c_phi_cond = estimator.c_phi(cond_sample) + cond_cost

                # Add zeros to ensure plates align
                c_phi_relaxed = torch.zeros_like(cost)
                c_phi_relaxed[_index1] = _c_phi_relaxed
                c_phi_cond = torch.zeros_like(cost)
                c_phi_cond[_index1] = _c_phi_cond
                return c_phi_relaxed, c_phi_cond

            return (log_prob,) + compute_REBAR_cvs(self, tensor, cost)
        else:
            hard_sample = discretize(tensor, distribution)
            relaxed_sample = tensor
            cond_sample = storch.conditional_gumbel_rsample(
                hard_sample, distribution.probs, isinstance(distribution, torch.distributions.Bernoulli), self.temperature
            )
            # Input rsampled values into c_phi
            _c_phi_relaxed = self.c_phi(relaxed_sample)
            _c_phi_cond = self.c_phi(cond_sample)
            _log_prob = distribution.log_prob(hard_sample)
            return _log_prob, _c_phi_relaxed, _c_phi_cond

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        # TODO: This estimator likely doesn't reduce plate weightings properly

        plate = tensor.get_plate(tensor.name)
        plate_index = -1
        if plate.n > 1:
            plate_index = tensor.get_plate_dim_index(plate.name)

        log_prob, c_phi_z, c_phi_tilde_z = self.compute_estimator(
            tensor,
            cost_node,
            plate_index,
            plate.n // 3 if self.rebar else plate.n,
            tensor.distribution,
        )
        c_phi_z = self.eta * c_phi_z
        c_phi_tilde_z = self.eta * c_phi_tilde_z
        log_prob = log_prob.sum(log_prob.event_dim_indices)
        return (
            log_prob,
            c_phi_z - c_phi_z.detach()
            - c_phi_tilde_z + (2-magic_box(log_prob)) * c_phi_tilde_z.detach()
        )

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor)]
    ) -> None:
        # During the normal backwards call, the parameters are accumulated gradients to the control variate parameters.
        # We don't want to minimize wrt to that loss, but to the one we define here.
        for param in self.control_params:
            param.grad = None
        # Find grad of the distribution, then compute variance.
        # TODO: check if the set statement properly filters different results here
        tensors = list(set([result[0] for result in result_triples]))
        # minimize the variance of the gradient with respect to the input parameters
        for tensor in tensors:
            # TODO: We have to select the probs of the distribution here as that's what it flows to. Is this always correct?
            d_param = tensor.grad['probs']
            variance = (d_param ** 2).sum(d_param.event_dim_indices)
            var_loss = storch.reduce_plates(variance)

            d_variance = torch.autograd.grad(
                [var_loss._tensor], self.control_params, retain_graph=True, allow_unused=True
            )
            for i in range(len(self.control_params)):
                self.control_params[i].backward(d_variance[i])

    def should_create_higher_order_graph(self) -> bool:
        return True


class REBAR(RELAX):
    """
    Gradient estimator for Bernoulli and Categorical distributions on differentiable functions.
    Implements REBAR as a special case of RELAX.
    See Tucker et al, 2017, https://arxiv.org/abs/1703.07370
    """

    def __init__(
        self,
        plate_name: str,
        *,
        c_phi: Callable[[torch.Tensor], torch.Tensor] = None,
        n_samples: int = 1,
    ):
        # Default REBAR does not use external control variate neural network c_phi
        if not c_phi:
            c_phi = lambda x: 0.0
        super().__init__(plate_name, n_samples=n_samples, c_phi=c_phi, rebar=True)



from .wrappers import (
    deterministic,
    stochastic,
    reduce,
    _exception_wrapper,
    _unpack_wrapper,
    ignore_wrapping,
)
from .tensor import Tensor, CostTensor, StochasticTensor, Plate, is_tensor

import storch.sampling
from .storch import *
import storch.method
import storch.typing
from .inference import backward, add_cost, reset, denote_independent, gather_samples
from .util import print_graph

from .unique import unique, undo_unique
from .exceptions import IllegalStorchExposeError
from packaging import version

import storch.nn


import torch as _torch
import pkgutil
import sys

_debug = False
# Hard-coded monkey patches: These do not support __torch_function__ (PyTorch version 1.5.0)
_torch.is_tensor = storch.is_tensor
_torch.Tensor.to = deterministic(_torch.Tensor.to)
_torch.Tensor.le = deterministic(_torch.Tensor.le)
_torch.Tensor.gt = deterministic(_torch.Tensor.gt)
# Cat currently does not support __torch_function__ https://github.com/pytorch/pytorch/issues/34294.
# However, monkey patching it makes storchastic incompatible with torchvision. Use storch.cat or storch.gather_samples instead
# _torch.cat = deterministic(_torch.cat)

# broadcast_all is not compatible with Tensor-likes... But a lot of Distributions code depends on it.
# Monkey patch every occurence of broadcast_all in the PyTorch code.
# This should not be necessesary from PyTorch 1.8.0. See https://github.com/pytorch/pytorch/pull/48169
if version.parse(torch.__version__) < version.parse("1.8.0"):
    _torch.distributions.utils.broadcast_all = deterministic(
        _torch.distributions.utils.broadcast_all
    )

# Distributions import broadcast_all by name, meaning they refer to the non-monkey patched version.
# Loop over every distribution to apply the monkey patch.
for module_info in pkgutil.iter_modules(_torch.distributions.__path__):
    module = sys.modules.get("torch.distributions." + module_info.name)
    if "broadcast_all" in module.__dict__:
        module.broadcast_all = _torch.distributions.utils.broadcast_all

# Validating arguments is not compatible with Storchastic as it does instanceof torch.Tensor checks
# Necessary for PyTorch versions above 1.8.0
_torch.distributions.Distribution.set_default_validate_args(False)


import storch
import torch
from torch.distributions import Bernoulli, OneHotCategorical
from storch.method import RELAX, REBAR, ARM

torch.manual_seed(0)

p = torch.tensor(0.5, requires_grad=True)
d = Bernoulli(p)
# sample = RELAX("sample", in_dim=1)(d)
sample = ARM('sample', n_samples=10)(d)
storch.add_cost(sample, "cost")
storch.backward()

method = REBAR("test", n_samples=1)
x = torch.Tensor([[0.2, 0.4, 0.4], [0.5, 0.1, 0.4], [0.2, 0.2, 0.6], [0.15, 0.15, 0.7]])
qx = OneHotCategorical(x)
print(method(qx))


from __future__ import annotations

import itertools
from abc import abstractmethod
from typing import Union, List, Optional, Tuple

from storch.typing import AnyTensor
from storch.sampling.method import SamplingMethod
from torch.distributions import Distribution
import storch
import torch


class AncestralPlate(storch.Plate):
    def __init__(
        self,
        name: str,
        n: int,
        parents: List[storch.Plate],
        variable_index: int,
        parent_plate: AncestralPlate,
        selected_samples: storch.Tensor,
        log_probs: storch.Tensor,
        weight: Optional[storch.Tensor] = None,
    ):
        super().__init__(name, n, parents, weight)
        assert (not parent_plate and variable_index == 0) or (
            parent_plate.n <= self.n and parent_plate.variable_index < variable_index
        )
        self.parent_plate = parent_plate
        self.selected_samples = selected_samples
        self.log_probs = storch.Tensor(
            log_probs._tensor, [log_probs], log_probs.plates + [self]
        )
        self.variable_index = variable_index
        self._in_recursion = False
        self._override_equality = False

    def __eq__(self, other):
        if self._override_equality:
            return other.name == self.name
        return (
            super().__eq__(other)
            and isinstance(other, AncestralPlate)
            and self.variable_index == other.variable_index
        )

    def __repr__(self):
        return (
            "(Ancestral, " + self.variable_index.__repr__() + super().__repr__() + ")"
        )

    def on_collecting_args(self, plates: [storch.Plate]) -> bool:
        """
        Filter the collected plates to only keep the AncestralPlates (with the same name) that has the highest variable index.
        :param plates:
        :return:
        """
        if self._in_recursion:
            self._override_equality = True
        return super().on_collecting_args(plates)

    def on_duplicate_plate(self, plate: storch.Plate) -> bool:
        if not isinstance(plate, AncestralPlate):
            raise ValueError(
                "Received a plate with name "
                + plate.name
                + " that is not also an AncestralPlate."
            )
        return plate.variable_index > self.variable_index

    def on_unwrap_tensor(self, tensor: storch.Tensor) -> storch.Tensor:
        """
        Gets called whenever the given tensor is being unwrapped and unsqueezed for batch use.
        This method should not be called on tensors whose variable index is higher than this plates.

        :param tensor: The input tensor that is being unwrapped
        :return: The tensor that will be unwrapped and unsqueezed in the future. Can be a modification of the input tensor.
        """
        if self._in_recursion:
            # Required when calling storch.gather in this method. It will call on_unwrap_tensor again.
            return tensor
        for i, plate in enumerate(tensor.multi_dim_plates()):
            if plate.name != self.name:
                continue
            assert isinstance(plate, AncestralPlate)
            if plate.variable_index == self.variable_index:
                return tensor
            # This is true by the filtering at on_collecting_args
            assert plate.variable_index < self.variable_index

            parent_plates = []
            current_plate = self

            # Collect the list of plates from the tensors variable index to this plates variable index
            while current_plate.variable_index != plate.variable_index:
                parent_plates.append(current_plate)
                current_plate = current_plate.parent_plate
            assert current_plate == plate

            # Go over all parent plates and gather their respective choices.
            for parent_plate in reversed(parent_plates):
                self._in_recursion = True
                expanded_selected_samples = expand_with_ignore_as(
                    parent_plate.selected_samples, tensor, self.name
                )
                self._override_equality = False
                # Gather what samples of the tensor are chosen by this plate (parent_plate)
                tensor = storch.gather(
                    tensor, parent_plate.name, expanded_selected_samples
                )
                self._in_recursion = False
                self._override_equality = False
            break
        return tensor


class SequenceDecoding(SamplingMethod):
    """
    Methods for generating sequences of discrete random variables.
    Examples: Simple ancestral sampling with replacement, beam search, Stochastic beam search (sampling without replacement)
    """

    EPS = 1e-8

    def __init__(self, plate_name: str, k: int, eos: None):
        super().__init__(plate_name)
        self.k = k
        self.eos = eos
        self.reset()

    def reset(self):
        super().reset()
        # Cumulative log probabilities of the samples
        self.joint_log_probs = None
        # Chosen parent samples at the previous sample step
        self.parent_indexing = None
        # The index of the currently sampled variable
        self.variable_index = 0
        # The previously created plates
        self.new_plate = None
        # What runs are already finished
        self.finished_samples = None
        # The sampled sequence so far
        self.seq = []

    def sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        orig_distr_plates: [storch.Plate],
        requires_grad: bool,
    ) -> (torch.Tensor, storch.Plate):

        """
        Sample from the distribution given the sequence so far.
        :param distribution: The distribution to sample from
        :return:
        """

        # This code has three parts.
        # The first prepares all necessary tensors to make sure they can be easily indexed.
        # This part is quite long as there are many cases.
        # 1) There have not been any variables sampled so far.
        # 2) There have been variables sampled, but their results are NOT used to compute the input distribution.
        #    in other words, the variable to sample is independent of the other sampled variables. However,
        #    we should still keep track of the other sampled variables to make sure that it still samples without
        #    replacement properly. In this case, the ancestral plate is not in the plates attribute.
        #    We also have to make sure that we know in the future what samples are chosen for the _other_ samples.
        # 3) There have been parents sampled, and this variable is dependent on at least some of them.
        #    The plates list then contains the ancestral plate. We need to make sure we compute the joint log probs
        #    for the conditional samples (ie, based on the different sampled variables in the ancestral dimension).
        # The second part is a loop over all options for the event dimensions. This samples these conditionally
        # independent samples in sequence. It samples indexes, not events.
        # The third part after the loop uses the sampled indexes and matches it to the events to be used.

        # LEGEND FOR SHAPE COMMENTS
        # =========================
        # To make this code generalize to every bayesian network, complicated shape management is necessary.
        # The following are references to the shapes that are used within the method
        #
        # distr_plates: refers to the plates on the parameters of the distribution. Does *not* include
        #  the k? ancestral plate (possibly empty)
        # orig_distr_plates: refers to the plates on the parameters of the distribution, and *can* include
        #  the k? ancestral plate (possibly empty)
        # prev_plates: refers to the plates of the previous sampled variable in this swr sample (possibly empty)
        # plates: refers to all plates except this ancestral plate, of which there are amt_plates. The first plates are
        #  the distr_plates, after that the prev_plates that are _not_ in distr_plates.
        #  It is composed of distr_plate x (ancstr_plates - distr_plates)
        # events: refers to the conditionally independent dimensions of the distribution (the distributions batch shape minus the plates)
        # k: refers to self.k
        # k?: refers to an optional plate dimension of this ancestral plate. It either doesn't exist, or is the sample
        #  dimension. If it exists, this means this sample is conditionally dependent on ancestors.
        # |D_yv|: refers to the *size* of the domain
        # amt_samples: refers to the current amount of sampled sequences. amt_samples <= k, but it can be lower if there
        #  are not enough events to sample from (eg |D_yv| < k)
        # event_shape: refers to the *shape* of the domain elements
        #  (can be 0, eg Categorical, or equal to |D_yv| for OneHotCategorical)

        # Do the decoding step given the prepared tensors
        samples, self.joint_log_probs, self.parent_indexing = self.decode(
            distr, self.joint_log_probs, parents, orig_distr_plates
        )

        # Find out what sequences have reached the EOS token, and make sure to always sample EOS after that.
        # Does not contain the ancestral plate as this uses samples instead of s_tensor.
        if self.eos:
            self.finished_samples = samples.eq(self.eos)

        k_index = 0
        plates = orig_distr_plates
        if isinstance(samples, storch.Tensor):
            k_index = samples.plate_dims
            plates = samples.plates
            samples = samples._tensor

        plate_size = samples.shape[k_index]

        # Remove the ancestral plate, if it already happens to be in
        to_remove = None
        for plate in plates:
            if plate.name == self.plate_name:
                to_remove = plate
                break
        if to_remove:
            plates.remove(to_remove)

        # Create the newly updated plate
        self.new_plate = self.create_plate(plate_size, plates.copy())
        plates.append(self.new_plate)

        if self.parent_indexing is not None:
            self.parent_indexing.plates.append(self.new_plate)

        self.seq = list(map(lambda t: self.new_plate.on_unwrap_tensor(t), self.seq))

        # Construct the stochastic tensor
        s_tensor = storch.StochasticTensor(
            samples,
            parents,
            plates,
            self.plate_name,
            plate_size,
            distr,
            requires_grad or self.joint_log_probs.requires_grad,
        )

        self.seq.append(s_tensor)

        # Increase variable index
        self.variable_index += 1
        return s_tensor, self.new_plate

    def weighting_function(
        self, tensor: storch.StochasticTensor, plate: storch.Plate
    ) -> Optional[storch.Tensor]:
        if self.eos:
            active = 1 - self.finished_samples
            amt_active: storch.Tensor = storch.sum(active, plate)
            return active / amt_active
        return super().weighting_function(tensor, plate)

    @abstractmethod
    def decode(
        self,
        distribution: Distribution,
        joint_log_probs: Optional[storch.Tensor],
        parents: [storch.Tensor],
        orig_distr_plates: [storch.Plate],
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor):
        """
        Decode given the input arguments
        :param distribution: The distribution to decode
        :param joint_log_probs: The log probabilities of the samples so far. prev_plates x amt_samples
        :param parents: List of parents of this tensor
        :param orig_distr_plates: List of plates from the distribution. Can include the self plate k.
        :return: 3-tuple of `storch.Tensor`. 1: The sampled value. 2: The new joint log probabilities of the samples.
        3: How the samples index the parent samples. Can just be a range if there is no choosing happening.
        For all of these, the last plate index should be the plate index, with the other plates like `all_plates`
        """
        pass

    def create_plate(self, plate_size: int, plates: [storch.Plate]) -> AncestralPlate:
        return AncestralPlate(
            self.plate_name,
            plate_size,
            plates.copy(),
            self.variable_index,
            self.new_plate,
            self.parent_indexing,
            self.joint_log_probs,
            None,
        )

    def get_sampled_seq(self, finished: bool = False) -> [storch.StochasticTensor]:
        # TODO: the finished one doesn't return proper tensors.
        if finished:
            return list(map(lambda t: t[self.finished_samples], self.seq))
        return torch.cat(self.seq, dim=self.seq[0].plate_dims)

    def get_amt_finished(self) -> AnyTensor:
        if not self.eos:
            raise RuntimeError(
                "Cannot get the amount of finished sequences when eos is not set."
            )
        if self.finished_samples is None:
            return 0
        return torch.sum(self.finished_samples, -1)

    def get_unique_seqs(self):
        # TODO: Very experimental code
        seq_dim = self.seq[0].plate_dims
        cat_seq = torch.cat(self.seq, dim=seq_dim)
        return storch.unique(cat_seq, event_dim=0)

    def all_finished(self) -> bool:
        t = self.get_amt_finished().eq(self.k)
        # This is required because storch.Tensor's do not support .all() and .bool()
        if isinstance(t, storch.Tensor):
            return t._tensor.all().bool()
        return t.all().bool()


class MCDecoder(SequenceDecoding):
    def decode(
        self,
        distribution: Distribution,
        joint_log_probs: Optional[storch.Tensor],
        parents: [storch.Tensor],
        orig_distr_plates: [storch.Plate],
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor):

        is_conditional_sample = False

        for plate in orig_distr_plates:
            if plate.name == self.plate_name:
                is_conditional_sample = True

        if is_conditional_sample:
            sample = self.mc_sample(
                distribution, parents, orig_distr_plates, 1
            ).squeeze(0)
        else:
            sample = self.mc_sample(distribution, parents, orig_distr_plates, self.k)

        s_log_probs = distribution.log_prob(sample)
        if joint_log_probs:
            joint_log_probs += s_log_probs
        else:
            joint_log_probs = s_log_probs

        if self.finished_samples:
            # Make sure we do not change the log probabilities for samples that were already finished
            joint_log_probs = (
                joint_log_probs * self.finished_samples
                + joint_log_probs * (1 - self.finished_samples)
            )
            sample[self.finished_samples] = self.eos

        return sample, joint_log_probs, None


class IterDecoding(SequenceDecoding):
    def decode(
        self,
        distr: Distribution,
        joint_log_probs: Optional[storch.Tensor],
        parents: [storch.Tensor],
        orig_distr_plates: [storch.Plate],
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor):
        """
        Decode given the input arguments
        :param distribution: The distribution to decode
        :param joint_log_probs: The log probabilities of the samples so far. prev_plates x amt_samples
        :param parents: List of parents of this tensor
        :param orig_distr_plates: List of plates from the distribution. Can include the self plate k.
        :return: 3-tuple of `storch.Tensor`. 1: The sampled value. 2: The new joint log probabilities of the samples.
        3: How the samples index the parent samples. Can just be a range if there is no choosing happening.
        """
        ancestral_distrplate_index = -1
        is_conditional_sample = False

        multi_dim_distr_plates = []
        multi_dim_index = 0
        for plate in orig_distr_plates:
            if plate.n > 1:
                if plate.name == self.plate_name:
                    ancestral_distrplate_index = multi_dim_index
                    is_conditional_sample = True
                else:
                    multi_dim_distr_plates.append(plate)
                multi_dim_index += 1
        # plates? x k x events x

        # TODO: This doesn't properly combine two ancestral plates with the same name but different variable index
        #  (they should merge).
        all_multi_dim_plates = multi_dim_distr_plates.copy()
        if self.variable_index > 0:
            # Previous variables have been sampled. add the prev_plates to all_plates
            for plate in self.joint_log_probs.multi_dim_plates():
                if plate not in multi_dim_distr_plates:
                    all_multi_dim_plates.append(plate)

        amt_multi_dim_plates = len(all_multi_dim_plates)
        amt_multi_dim_distr_plates = len(multi_dim_distr_plates)
        amt_multi_dim_orig_distr_plates = amt_multi_dim_distr_plates + (
            1 if is_conditional_sample else 0
        )
        amt_multi_dim_prev_plates = amt_multi_dim_plates - amt_multi_dim_distr_plates
        if not distr.has_enumerate_support:
            raise ValueError("Can only decode distributions with enumerable support.")

        with storch.ignore_wrapping():
            # |D_yv| x (|distr_plates| + |k?| + |event_dims|) * (1,) x |D_yv|
            support_non_expanded: torch.Tensor = distr.enumerate_support(expand=False)
            # Compute the log-probability of the different events
            # |D_yv| x distr_plate[0] x ... k? ... x distr_plate[n-1] x events
            d_log_probs = distr.log_prob(support_non_expanded)

            # Note: Use amt_orig_distr_plates here because it might include k? dimension. amt_distr_plates filters this one.
            # distr_plate[0] x ... k? ... x distr_plate[n-1] x |D_yv| x events
            d_log_probs = storch.Tensor(
                d_log_probs.permute(
                    tuple(range(1, amt_multi_dim_orig_distr_plates + 1))
                    + (0,)
                    + tuple(
                        range(
                            amt_multi_dim_orig_distr_plates + 1, len(d_log_probs.shape)
                        )
                    )
                ),
                [],
                orig_distr_plates,
            )

        # |D_yv| x distr_plate[0] x ... x k? x ... x distr_plate[n-1] x events x event_shape
        support = distr.enumerate_support(expand=True)

        if is_conditional_sample:
            # Reduce ancestral dimension in the support. As the dimension is just an expanded version, this should
            # not change the underlying data.
            # |D_yv| x distr_plates x events x event_shape
            support = support[(slice(None),) * (ancestral_distrplate_index + 1) + (0,)]

            # Gather the correct log probabilities
            # distr_plate[0] x ... k ... x distr_plate[n-1] x |D_yv| x events
            # TODO: Move this down below to the other scary TODO
            d_log_probs = self.new_plate.on_unwrap_tensor(d_log_probs)
            # Permute the dimensions of d_log_probs st the k dimension is after the plates.
            for i, plate in enumerate(d_log_probs.multi_dim_plates()):
                if plate.name == self.plate_name:
                    d_log_probs.plates.remove(plate)
                    # distr_plates x k x |D_yv| x events
                    d_log_probs._tensor = d_log_probs._tensor.permute(
                        tuple(range(0, i))
                        + tuple(range(i + 1, amt_multi_dim_orig_distr_plates))
                        + (i,)
                        + tuple(
                            range(
                                amt_multi_dim_orig_distr_plates, len(d_log_probs.shape)
                            )
                        )
                    )
                    break

        # Equal to event_shape
        element_shape = distr.event_shape
        support_permutation = (
            tuple(range(1, amt_multi_dim_distr_plates + 1))
            + (0,)
            + tuple(range(amt_multi_dim_distr_plates + 1, len(support.shape)))
        )
        # distr_plates x |D_yv| x events x event_shape
        support = support.permute(support_permutation)

        if amt_multi_dim_plates != amt_multi_dim_distr_plates:
            # If previous samples had plate dimensions that are not in the distribution plates, add these to the support.
            support = support[
                (slice(None),) * amt_multi_dim_distr_plates
                + (None,) * amt_multi_dim_prev_plates
            ]
            all_plate_dims = tuple(map(lambda _p: _p.n, all_multi_dim_plates))
            # plates x |D_yv| x events x event_shape (where plates = distr_plates x prev_plates)
            support = support.expand(
                all_plate_dims + (-1,) * (len(support.shape) - amt_multi_dim_plates)
            )
        # plates x |D_yv| x events x event_shape
        support = storch.Tensor(support, [], all_multi_dim_plates)

        # Equal to events: Shape for the different conditional independent dimensions
        event_shape = support.shape[
            amt_multi_dim_plates + 1 : -len(element_shape)
            if len(element_shape) > 0
            else None
        ]

        ranges = []
        for size in event_shape:
            ranges.append(list(range(size)))

        amt_samples = 0
        parent_indexing = None
        if joint_log_probs is not None:
            # Initialize a tensor (self.parent_indexing) that keeps track of what samples link to previous choices of samples
            # Note that joint_log_probs.shape[-1] is amt_samples, not k. It's possible that amt_samples < k!
            amt_samples = joint_log_probs.shape[-1]
            # plates x k
            parent_indexing = support.new_zeros(
                size=support.shape[:amt_multi_dim_plates] + (self.k,), dtype=torch.long
            )

            # probably can go wrong if plates are missing.
            parent_indexing[..., :amt_samples] = left_expand_as(
                torch.arange(amt_samples), parent_indexing
            )
        # plates x k x events
        sampled_support_indices = support.new_zeros(
            size=support.shape[:amt_multi_dim_plates]  # plates
            + (self.k,)
            + support.shape[
                amt_multi_dim_plates + 1 : -len(element_shape)
                if len(element_shape) > 0
                else None
            ],  # events
            dtype=torch.long,
        )
        # Sample independent tensors in sequence
        # Iterate over the different (conditionally) independent samples being taken (the events)
        for indices in itertools.product(*ranges):
            # Log probabilities of the different options for this sample step (event)
            # distr_plates x k? x |D_yv|
            yv_log_probs = d_log_probs[(...,) + indices]
            (
                sampled_support_indices,
                joint_log_probs,
                parent_indexing,
                amt_samples,
            ) = self.decode_step(
                indices,
                yv_log_probs,
                joint_log_probs,
                sampled_support_indices,
                parent_indexing,
                is_conditional_sample,
                amt_multi_dim_plates,
                amt_samples,
            )
        # Finally, index the support using the sampled indices to get the sample!
        if amt_samples < self.k:
            # plates x amt_samples x events
            sampled_support_indices = sampled_support_indices[
                (...,) + (slice(amt_samples),) + (slice(None),) * len(ranges)
            ]
        expanded_indices = right_expand_as(sampled_support_indices, support)
        sample = support.gather(dim=amt_multi_dim_plates, index=expanded_indices)
        return sample, joint_log_probs, parent_indexing

    @abstractmethod
    def decode_step(
        self,
        indices: Tuple[int],
        yv_log_probs: storch.Tensor,
        joint_log_probs: Optional[storch.Tensor],
        sampled_support_indices: Optional[storch.Tensor],
        parent_indexing: Optional[storch.Tensor],
        is_conditional_sample: bool,
        amt_plates: int,
        amt_samples: int,
    ) -> (storch.Tensor, storch.Tensor, storch.Tensor, int):
        """
        Decode given the input arguments for a specific event
        :param indices: Tuple of integers indexing the current event to sample.
        :param yv_log_probs:  Log probabilities of the different options for this event. distr_plates x k? x |D_yv|
        :param joint_log_probs: The log probabilities of the samples so far. None if `not is_conditional_sample`. prev_plates x amt_samples
        :param sampled_support_indices: Tensor of samples so far. None if this is the first set of indices. plates x k x events
        :param parent_indexing: Tensor indexing the parent sample. None if `not is_conditional_sample`.
        :param is_conditional_sample: True if a parent has already been sampled. This means the plates are more complex!
        :param amt_plates: The total amount of plates in both the distribution and the previously sampled variables
        :param amt_samples: The amount of active samples.
        :return: 3-tuple of `storch.Tensor`. 1: sampled_support_indices, with `:, indices` referring to the indices for the support.
        2: The updated `joint_log_probs` of the samples.
        3: The updated `parent_indexing`. How the samples index the parent samples. Can just return parent_indexing if nothing happens.
        4: The amount of active samples after this step.
        """
        pass


def expand_with_ignore_as(
    tensor, expand_as, ignore_dim: Union[str, int]
) -> torch.Tensor:
    """
    Expands the tensor like expand_as, but ignores a single dimension.
    Ie, if tensor is of size a x b,  expand_as of size d x a x c and dim=-1, then the return will be of size d x a x b.
    It also automatically expands all plate dimensions correctly.
    :param ignore_dim: Can be a string referring to the plate dimension
    """
    # diff = expand_as.ndim - tensor.ndim
    def _expand_with_ignore(tensor, expand_as, dim: int):
        new_dims = expand_as.ndim - tensor.ndim
        # after_dims = tensor.ndim - dim
        return tensor[(...,) + (None,) * new_dims].expand(
            expand_as.shape[:dim]
            + (-1,)
            + (expand_as.shape[dim + 1 :] if dim != -1 else ())
        )

    if isinstance(ignore_dim, str):
        return storch.deterministic(
            _expand_with_ignore, expand_plates=True, dim=ignore_dim
        )(tensor, expand_as)
    return storch.deterministic(_expand_with_ignore, expand_plates=True)(
        tensor, expand_as, ignore_dim
    )


@storch.deterministic(l_broadcast=False)
def right_expand_as(tensor, expand_as):
    diff = expand_as.ndim - tensor.ndim
    return tensor[(...,) + (None,) * diff].expand(
        (-1,) * tensor.ndim + expand_as.shape[tensor.ndim :]
    )


def left_expand_as(tensor, expand_as):
    diff = expand_as.ndim - tensor.ndim
    return tensor[(None,) * diff].expand(expand_as.shape[:diff] + (-1,) * tensor.ndim)



import torch
from torch.distributions import Normal
from storch.method import Reparameterization, ScoreFunction
import storch

torch.manual_seed(0)


def compute_f(method):
    a = torch.tensor(5.0, requires_grad=True)
    b = torch.tensor(-3.0, requires_grad=True)
    c = torch.tensor(0.23, requires_grad=True)
    d = a + b

    # Sample e from a normal distribution using reparameterization
    normal_distribution = Normal(b + c, 1)
    e = method(normal_distribution)

    f = d * e * e
    return f, c


# e*e follows a noncentral chi-squared distribution https://en.wikipedia.org/wiki/Noncentral_chi-squared_distribution
# exp_f = d * (1 + mu * mu)
repar = Reparameterization("e", n_samples=1)
f, c = compute_f(repar)
storch.add_cost(f, "f")
print(storch.backward())

print("first derivative estimate", c.grad)

f, c = compute_f(repar)
storch.add_cost(f, "f")
print(storch.backward())

print("second derivative estimate", c.grad)


def estimate_variance(method):
    gradient_samples = []
    for i in range(1000):
        f, c = compute_f(method)
        storch.add_cost(f, "f")
        storch.backward()
        gradient_samples.append(c.grad)
    gradients = storch.gather_samples(gradient_samples, "gradients")
    # print(gradients)
    print("variance", storch.variance(gradients, "gradients"))
    print("mean", storch.reduce_plates(gradients, "gradients"))
    print("st dev", torch.sqrt(storch.variance(gradients, "gradients")))

    print(type(gradients))
    print(gradients.shape)
    print(gradients.plates)


print("Reparameterization n=1")
estimate_variance(Reparameterization("e", n_samples=1))

print("Reparameterization n=10")
estimate_variance(Reparameterization("e", n_samples=10))

print("Score function n=1")
estimate_variance(ScoreFunction("e", n_samples=1))

print("Score function n=45")
estimate_variance(ScoreFunction("e", n_samples=45))

print("Score function with baseline n=20")
estimate_variance(ScoreFunction("e", n_samples=20, baseline_factory="batch_average"))



from examples.vae import VAE, main
import torch.nn as nn
import torch
import storch
from torch.distributions import Distribution, Bernoulli
from storch.method import (
    ScoreFunctionWOR,
    UnorderedSetEstimator,
    RELAX,
    REBAR,
    ScoreFunction,
    GumbelSoftmax,
    Expect,
)
from torch.distributions.utils import clamp_probs


class BernoulliVAE(VAE):
    def initialize_param_layers(
        self, latents: int, prev_layer: int
    ) -> (nn.Module, nn.Module):
        fc3 = nn.Linear(prev_layer, latents)
        fc4 = nn.Linear(latents, prev_layer)
        return fc3, fc4

    def initialize_method(self, args) -> storch.method.Method:
        if args.method == "gumbel":
            return GumbelSoftmax("z", n_samples=args.samples)
        if args.method == "gumbel_sparse":
            return storch.method.GumbelSparseMax("z", n_samples=args.samples)
        if args.method == "gumbel_entmax":
            return storch.method.GumbelEntmax(
                "z", n_samples=args.samples, adaptive=True
            )
        elif args.method == "gumbel_straight":
            return GumbelSoftmax("z", n_samples=args.samples, straight_through=True)
        elif args.method == "gumbel_wor":
            return storch.method.UnorderedSetGumbelSoftmax("z", k=args.samples)
        elif args.method == "score":
            return ScoreFunction(
                "z", n_samples=args.samples, baseline_factory=args.baseline
            )
        elif args.method == "expect":
            return Expect("z")
        elif args.method == "relax":
            return RELAX("z", n_samples=args.samples, in_dim=(args.latents,))
        elif args.method == "rebar":
            return REBAR("z", n_samples=args.samples)
        elif args.method == "relax_rebar":
            return RELAX(
                "z", n_samples=args.samples, in_dim=(args.latents,), rebar=True
            )
        elif args.method == "score_wor":
            return ScoreFunctionWOR("z", k=args.samples)
        elif args.method == "unordered":
            return UnorderedSetEstimator("z", k=args.samples)
        elif args.method == "arm":
            return storch.method.ARM("z", n_samples=args.samples)
        elif args.method == "disarm":
            return storch.method.DisARM("z", n_samples=args.samples)
        else:
            raise ValueError("Invalid method passed to program arguments.")

    def prior(self, posterior: Distribution):
        return Bernoulli(probs=torch.ones_like(posterior.probs) / 2.0)

    def variational_posterior(self, logits: torch.Tensor):
        return Bernoulli(probs=clamp_probs(logits.sigmoid()))

    def logits_to_params(self, logits, latents):
        return logits

    def shape_latent(self, z, latents):
        if z.dtype == torch.bool:
            return z.float()
        return z

    def name(self):
        return "bernoulli vae"


if __name__ == "__main__":
    main(BernoulliVAE)



import storch
import torch
import torch.distributions as td
method1 = storch.method.Reparameterization
method2 = storch.method.ScoreFunction

method1 = method1(plate_name="1",n_samples=25)
method2 = method2(plate_name="1",n_samples=25)
p1 = td.Independent(td.Normal(loc=torch.zeros([1000, 2]), scale=torch.ones([1000, 2])), 0)
p2 = td.Independent(td.OneHotCategorical(probs=torch.zeros([1000, 3]).uniform_()), 0)

samp1 = method1(p1)
samp2 = method2 (p2)
# torch.Size([25, 1000, 2])
# torch.Size([25, 1000, 3])

print(storch.cat([samp1,samp2], 2).shape)
# torch.Size([25, 1000, 5])

method1 = storch.method.Reparameterization
method2 = storch.method.UnorderedSetEstimator

method1 = method1(plate_name="1",n_samples=25)
method2 = method2(plate_name="2",k=25)
p1 = td.Independent(td.Normal(loc=torch.zeros([1000, 2]), scale=torch.ones([1000, 2])), 0)
p2 = td.Independent(td.OneHotCategorical(probs=torch.zeros([1000, 3]).uniform_()), 0)

samp1 = method1(p1)
samp2 = method2 (p2)
# torch.Size([25, 1000, 2])
# torch.Size([25, 1000, 3])

print(storch.cat([samp1,samp2], -1).shape)
# torch.Size([25, 25, 1000, 5])


from storch.sampling.method import (
    SamplingMethod,
    MonteCarlo,
)
from storch.sampling.expect import Enumerate
from storch.sampling.swor import SampleWithoutReplacement
from storch.sampling.unordered_set import UnorderedSet, GumbelSoftmaxWOR



from __future__ import annotations
from abc import ABC
from torch.distributions import Distribution, OneHotCategorical, Bernoulli, Categorical

from storch.tensor import CostTensor, StochasticTensor, Plate
import torch
from typing import Optional, Type, Union, Dict, List, Callable, Tuple
from storch.util import (
    has_differentiable_path,
    get_distr_parameters,
    rsample_gumbel_softmax,
    rsample_gumbel,
)
import storch
from storch.method.baseline import MovingAverageBaseline, BatchAverageBaseline, Baseline
from storch.sampling import (
    SamplingMethod,
    MonteCarlo,
    Enumerate,
)
import entmax


class Method(ABC, torch.nn.Module):
    """
    Base class of gradient estimation methods.

    A :class:`Method` is a :class:`torch.nn.Module`, and can therefore contain parameters to optimize.
    Calling them (:meth:`forward`) with a PyTorch distribution returns a sampled Tensor of type :class:`storch.StochasticTensor`
    from that distribution that will use the gradient estimator in the backward pass when :func:`storch.backward` is called.

    Args:
        plate_name (str): The name of the :class:`.Plate` that samples of this method will use.
        sampling_method (storch.sampling.SamplingMethod): The method to sample tensors with given an input distribution.
    """

    def __init__(self, plate_name: str, sampling_method: SamplingMethod):
        super().__init__()
        self._estimation_pairs = []
        self.register_buffer("iterations", torch.tensor(0, dtype=torch.long))
        self.plate_name = plate_name
        self.sampling_method = sampling_method
        if not self.sampling_method.plate_name == plate_name:
            raise ValueError(
                "The plate name of the sampling method and the storch method should match."
            )

    def forward(self, distr: Distribution) -> StochasticTensor:
        """
        Calls the sample method to sample from the given distribution
        :param torch.distribution.Distribution distr: The distribution to sample from.
        :return: The sampled tensor
        :rtype: storch.tensor.StochasticTensor
        """
        return self.sample(distr)

    @staticmethod
    def _create_hook(sample: StochasticTensor, name: str, plates: List[Plate]):
        accum_grads = sample.param_grads
        del sample  # Remove from hook closure for GC reasons

        def hook(*args: Tuple[any]):
            # For some reason, this args unpacking is required for compatbility with registring on a .grad_fn...?
            # TODO: I'm sure there could be something wrong here
            grad = args[-1]
            if isinstance(grad, tuple):
                grad = grad[0]

            if name in accum_grads:
                accum_grads[name] = storch.Tensor(
                    accum_grads[name]._tensor + grad, [], plates, name + "_grad"
                )
            else:
                accum_grads[name] = storch.Tensor(grad, [], plates, name + "_grad")

        # Extremely complex way to ensure the GC cleans up the backward graph
        # What this does is delete the accum_grads field after inference. This is needed to ensure the
        # hook is no longer in scope even after relevant tensors are out of scope.
        # clean_graph() is called from StochasticTensor._clean
        # See also https://github.com/pytorch/pytorch/issues/12863
        def clean_graph() -> None:
            try:
                nonlocal accum_grads
                del accum_grads
            except (UnboundLocalError, NameError):
                pass
        return hook, clean_graph

    def sample(self, distr: Distribution) -> storch.tensor.StochasticTensor:
        """
        Samples from the given distribution according to this Method's sampling scheme.
        :param distr:
        :return:
        """
        # Unwrap the distributions parameters
        params: Dict[str, storch.Tensor] = get_distr_parameters(
            distr, filter_requires_grad=False
        )
        parents: [torch.Tensor] = storch.wrappers._stochastic_parents.copy()
        # If we are in an @stochastic context, external plates might exist.
        plates: [Plate] = storch.wrappers._plate_links.copy()
        requires_grad = False
        for name, p in params.items():
            requires_grad = requires_grad or p.requires_grad
            if isinstance(p, storch.Tensor):
                parent_found = False
                for _p in parents:
                    if _p is p:
                        parent_found = True
                        break
                if not parent_found:
                    parents.append(p)
                # The sample should have the batch links of the parameters in the distribution, if present.
                for plate in p.plates:
                    if plate not in plates:
                        plates.append(plate)

        for plate in plates:
            if plate.name == self.plate_name:
                self.sampling_method.on_plate_already_present(plate)

        s_tensor: StochasticTensor
        # Sample from the proposal distribution
        s_tensor, plate = self.sampling_method(distr, parents, plates, requires_grad)

        s_tensor._set_method(self)

        batch_weighting = self.sampling_method.weighting_function(s_tensor, plate)
        if batch_weighting is not None:
            plate.weight = batch_weighting
            # TODO: I don't think this code should be here.
            # if isinstance(batch_weighting, storch.Tensor):
            #     batch_weighting.plates[0] = plate

        clean_hooks = []
        for name, param in params.items():
            # TODO: Possibly could find the wrong gradients here if multiple distributions use the same parameter?
            # This maybe requires copying the tensor hm...
            if param.requires_grad:
                hook_plates = []
                if isinstance(distr, OneHotCategorical) or isinstance(
                    distr, Categorical
                ):
                    if param is not distr._param:
                        continue
                    # We only care about the input parameter. Ie, it returns both probs and logits, but only
                    # the one the user used to create the Distribution is of interest.
                    # These distributions first normalize their logits/probs. This causes incorrect gradient statistics.
                    # By taking a step back in the computation graph, we retrieve the correct parameter.
                    if isinstance(param, storch.Tensor):
                        hook_plates = param.plates
                        param = param._tensor
                    to_hook = param.grad_fn.next_functions[0][0]
                elif isinstance(param, storch.Tensor):
                    hook_plates = param.plates
                    to_hook = param._tensor
                else:
                    to_hook = param
                hook, clean_hook = Method._create_hook(s_tensor, name, hook_plates)
                s_tensor._clean_hooks.append(clean_hook)
                to_hook.register_hook(hook)

        # Possibly change something in the tensor now that it is wrapped and registered in the graph.
        # Used for example to rsample in LAX to detach the tensor so that it won't record gradients
        # in the normal forward pass.
        # TODO: That should really be easier...
        edited_sample = self.post_sample(s_tensor)

        # Register this sampling method as being used in this iteration so that we can reset this method after the backward pass
        if self not in storch.inference._sampling_methods:
            storch.inference._sampling_methods.append(self)

        if edited_sample is not None:
            new_s_tensor = storch.tensor.StochasticTensor(
                edited_sample._tensor,
                [s_tensor],
                s_tensor.plates,
                s_tensor.name,
                s_tensor.n,
                s_tensor.distribution,
                False,
            )
            new_s_tensor.param_grads = s_tensor.param_grads
            return new_s_tensor
        return s_tensor

    def _estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        self._estimation_pairs.append((tensor, cost_node))
        return self.estimator(tensor, cost_node)

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        """
        Returns two terms that will be used for inferring higher-order gradient estimates.
        - The first return is the gradient function. It will be multiplied with the cost function.
          To get correct, unbiased estimates, the cost_node should not be involved in the computation.
          In REINFORCE-based methods, this is usually the score function.
          Methods that do not use a multiplicative estimator can return None.
        - The second return is the control variate, for instance a baseline in score-function based methods.

        It is also possible to directly do a backwards call in this method, but this will prevent correct computation of
        higher-order derivatives.

        Note that this method is only called if :meth:`adds_loss` returns True.

        Args:
            tensor (storch.StochasticTensor): The sampled tensor to find the surrogate loss for.
        """
        return None, None

    def _update_parameters(self: Method):
        self.iterations += 1
        self.update_parameters(self._estimation_pairs)
        self._estimation_pairs = []

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor)]
    ) -> None:
        """
        Update the (hyper)parameters of the estimation method. Note that this is not the parameters
        to be optimized!
        """
        pass

    def is_pathwise(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        """
        Returns true if the gradient estimator is pathwise, that is, no external loss functions need to be created
        to compute the correct gradient. This includes the pathwise derivative (Reparameterization) and the expectation.
        """
        return False

    def post_sample(self, tensor: storch.StochasticTensor) -> Optional[storch.Tensor]:
        return None

    def reset(self) -> None:
        """
        This function gets called whenever storchastic is reset. This is after storch.backward() or storch.reset() is
        called. Can be used to reset this methods state to some initial state that has to happen every iteration. 
        """
        self.sampling_method.reset()

    def should_create_higher_order_graph(self) -> bool:
        return False


class Infer(Method):
    """
    Method that automatically chooses reparameterization if it is available, and otherwise the score function
    with appropriate baseline (moving average if n_samples = 1, else batch average).
    """

    def __init__(
        self,
        plate_name: str,
        distribution_type: Type[Distribution],
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
    ):
        if distribution_type.has_rsample:
            _method = Reparameterization(plate_name, sampling_method, n_samples)
        elif issubclass(distribution_type, OneHotCategorical) or issubclass(
            distribution_type, Bernoulli
        ):
            _method = GumbelSoftmax(plate_name, sampling_method, n_samples)
        else:
            _method = ScoreFunction(
                plate_name,
                sampling_method,
                n_samples,
                baseline_factory="moving_average"
                if n_samples == 1
                else "batch_average",
            )
        super().__init__(plate_name, _method.sampling_method)
        self._score_method = ScoreFunction(plate_name, sampling_method, n_samples)
        self._method = _method

    def estimator(
        self, tensor: StochasticTensor, cost_node: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor],  Optional[storch.Tensor]
    ]:
        return self._score_method.estimator(tensor, cost_node)

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor, torch.Tensor)]
    ):
        self._method.update_parameters(result_triples)
        self._score_method.update_parameters(result_triples)

    def is_pathwise(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return True
        else:
            # No automatic baselines. Use the score function.
            return False


class Reparameterization(Method):
    """
    Can only be used for reparameterizable distributions and when the function to minimize is differentiable.
    Default option for reparameterizable distributions.
    """

    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(plate_name, sampling_method.set_mc_sample(self.reparam_sample))

    def is_pathwise(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return True
        raise ValueError(
            "There is no differentiable path between the cost tensor and the stochastic tensor. "
            "We cannot use reparameterization. Use a different gradient estimator, or make sure your"
            "code is differentiable."
        )

    def reparam_sample(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ):
        if not distr.has_rsample:
            raise ValueError(
                "The input distribution has not implemented rsample. If you use a discrete "
                "distribution, make sure to use eg GumbelSoftmax."
            )
        return distr.rsample((amt_samples,))


class GumbelSoftmax(Method):
    """
    Method that implements the Gumbel Softmax relaxation with temperature annealing.
    Can only be used to find the derivative when all paths to the cost nodes are differentiable.
    Introduced in https://arxiv.org/abs/1611.01144 and https://arxiv.org/abs/1611.00712
    """

    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
        straight_through=False,
        initial_temperature=1.0,
        min_temperature=1.0e-4,
        annealing_rate=1.0e-5,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(
            plate_name, sampling_method.set_mc_sample(self.sample_gumbel),
        )

        self.straight_through = straight_through
        self.register_buffer("temperature", torch.tensor(initial_temperature))
        self.register_buffer("annealing_rate", torch.tensor(annealing_rate))
        self.register_buffer("min_temperature", torch.tensor(min_temperature))

    def sample_gumbel(
        self,
        distr: Distribution,
        parents: [storch.Tensor],
        plates: [Plate],
        amt_samples: int,
    ):
        return rsample_gumbel_softmax(
            distr, amt_samples, self.temperature, self.straight_through
        )

    def is_pathwise(self, tensor: StochasticTensor, cost_node: CostTensor) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return True
        raise ValueError(
            "There is no differentiable path between the cost tensor and the stochastic tensor. "
            "We cannot use reparameterization. Use a different gradient estimator, or make sure your"
            "code is differentiable."
        )

    def update_parameters(
        self, result_triples: [(StochasticTensor, CostTensor)]
    ) -> None:
        if self.temperature > self.min_temperature:
            self.temperature = (1 - self.annealing_rate) * self.temperature


class GumbelEntmax(Method):
    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[storch.sampling.SamplingMethod] = None,
        alpha: float = 1.5,
        adaptive=False,
        n_samples: int = 1,
        straight_through=False,
        initial_temperature=1.0,
        min_temperature=1.0e-4,
        annealing_rate=0.0,
    ):
        if not sampling_method:
            sampling_method = storch.sampling.MonteCarlo(plate_name, n_samples)
        super().__init__(
            plate_name, sampling_method.set_mc_sample(self.sample_gumbel_entmax),
        )
        self.adaptive = adaptive
        self.straight_through = straight_through
        self.register_buffer("temperature", torch.tensor(initial_temperature))
        self.register_buffer("annealing_rate", torch.tensor(annealing_rate))
        self.register_buffer("min_temperature", torch.tensor(min_temperature))
        self.alpha = alpha
        if adaptive:
            self.alpha = torch.nn.Parameter(
                torch.tensor(self.alpha, requires_grad=True)
            )
        if not adaptive and alpha == 1.5:
            self.entmax = entmax.entmax15
        elif not adaptive and alpha == 2.0:
            self.entmax = entmax.sparsemax
        else:
            if adaptive:
                self.entmax = lambda x: entmax.entmax_bisect(
                    x, torch.nn.functional.softplus(self.alpha - 1) + 1
                )
            else:
                self.entmax = lambda x: entmax.entmax_bisect(x, self.alpha)

    def sample_gumbel_entmax(
        self,
        distr: torch.distributions.Distribution,
        parents: [storch.Tensor],
        plates: [storch.Plate],
        amt_samples: int,
    ):
        import random

        # if random.randint(0, 10) == 0:
        #     print(torch.nn.functional.softplus(self.alpha - 1) + 1)
        gumbels = rsample_gumbel(distr, amt_samples)
        res = self.entmax(gumbels / self.temperature)
        return res

    def is_pathwise(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> bool:
        if has_differentiable_path(cost_node, tensor):
            # There is a differentiable path, so we will just use reparameterization here.
            return True
        raise ValueError(
            "There is no differentiable path between the cost tensor and the stochastic tensor. "
            "We cannot use reparameterization. Use a different gradient estimator, or make sure your"
            "code is differentiable."
        )

    def update_parameters(
        self, result_triples: [(storch.StochasticTensor, storch.CostTensor)]
    ) -> None:
        if self.temperature > self.min_temperature:
            self.temperature = (1 - self.annealing_rate) * self.temperature


class GumbelSparseMax(GumbelEntmax):
    """
    Method that implements the Gumbel Sparsemax relaxation with temperature annealing.
    Can only be used to find the derivative when all paths to the cost nodes are differentiable.
    Introduced in Appendix of Gradient Estimation with Stochastic Softmax Tricks https://arxiv.org/abs/2006.08063
    """

    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[storch.sampling.SamplingMethod] = None,
        n_samples: int = 1,
        straight_through=False,
        initial_temperature=1.0,
        min_temperature=1.0e-4,
        annealing_rate=0.0,
    ):
        super().__init__(
            plate_name,
            sampling_method,
            2.0,
            n_samples=n_samples,
            straight_through=straight_through,
            initial_temperature=initial_temperature,
            min_temperature=min_temperature,
            annealing_rate=annealing_rate,
        )


BaselineFactory = Callable[[storch.StochasticTensor, storch.CostTensor], Baseline]


class ScoreFunction(Method):
    """
    The score function multiplies the loss function with a log p(z) term to estimate the gradients. It is always
    applicable, but has high variance without variance reduction techniques.

    Args:
        plate_name (str): The name of the :class:`.Plate` that samples of this method will use.
        sampling_method (storch.sampling.SamplingMethod): The method to sample tensors with given an input distribution.
          if not given, this defaults to simple Monte Carlo sampling.
        n_samples (int): How many samples to use. This parameter is only used when :arg:`sampling_method` is not passed.
          Defaults to 1.
        baseline_factory: The :class:`storch.method.baseline.Baseline` to use. This is passed as a string (options:
          batch_average, moving_average) or as a BaselineFactory.
    """

    def __init__(
        self,
        plate_name: str,
        sampling_method: Optional[SamplingMethod] = None,
        n_samples: int = 1,
        baseline_factory: Optional[Union[BaselineFactory, str]] = None,
        **kwargs,
    ):
        if not sampling_method:
            sampling_method = MonteCarlo(plate_name, n_samples)
        super().__init__(plate_name, sampling_method)
        self.baseline_factory: Optional[BaselineFactory] = baseline_factory
        if isinstance(baseline_factory, str):
            if baseline_factory == "moving_average":
                # Baseline per cost possible? So this lookup/buffer thing is not necessary
                self.baseline_factory = lambda s, c: MovingAverageBaseline(**kwargs)
            elif baseline_factory == "batch_average":
                if sampling_method.n_samples <= 1:
                    raise ValueError(
                        "Batch average can only be used for n_samples > 1. "
                    )
                self.baseline_factory = lambda s, c: BatchAverageBaseline()
            elif baseline_factory == "none" or baseline_factory == "None":
                self.baseline_factory = None
            else:
                raise ValueError("Invalid baseline name", baseline_factory)

    def estimator(
        self, tensor: StochasticTensor, cost: CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        log_prob = tensor.distribution.log_prob(tensor)
        if len(log_prob.shape) > tensor.plate_dims:
            # Sum out over the event shape
            log_prob = log_prob.sum(
                dim=list(range(tensor.plate_dims, len(log_prob.shape)))
            )
        if self.baseline_factory:
            baseline_name = "_b_" + tensor.name + "_" + cost.name
            if not hasattr(self, baseline_name):
                setattr(self, baseline_name, self.baseline_factory(tensor, cost))
            baseline = getattr(self, baseline_name)
            baseline = baseline.compute_baseline(tensor, cost)
            return log_prob, (1.0 - log_prob) * baseline
        return log_prob, None


class Expect(Method):
    def __init__(self, plate_name: str, budget=10000):
        super().__init__(plate_name, Enumerate(plate_name, budget))



from typing import Optional

import storch
import torch

"""
This module removes all plates from the tensor, and computes functions over the unique elements. Using clever bookkeeping,
we can retrieve the associated plates to retrieve the original plates.
This is useful for (sequences of) discrete output distributions that sample with replacement. We can greatly reduce
the computation of the deterministic functions (such as a deterministic reward) by only computing the function over
the unique elements.

Note: this code is highly experimental.
"""


class UniquePlate(storch.Plate):
    def __init__(
        self,
        name: str,
        amt_unique: int,
        shrunken_plates: [storch.Plate],
        inv_indexing: torch.Tensor,
    ):
        super().__init__(name, amt_unique, [])
        self.inv_indexing = inv_indexing
        self.shrunken_plates = shrunken_plates
        self.weight = None

    def reduce(self, unique_tensor: storch.Tensor, detach_weights=True):
        non_unique = self.undo_unique(unique_tensor)
        return storch.reduce_plates(
            non_unique, plates=self.shrunken_plates, detach_weights=detach_weights
        )

    def on_collecting_args(self, plates: [storch.Plate]) -> bool:
        for plate in plates:
            if self.has_shrunken(plate):
                raise ValueError(
                    "Cannot call deterministic wrapper with input tensors of which some are unique and "
                    "some are not. Use storch.undo_unique on the unique tensors before calling."
                )
        return True

    def has_shrunken(self, plate: storch.Plate) -> bool:
        for _plate in self.shrunken_plates:
            if plate == _plate:
                return True
            if isinstance(_plate, UniquePlate) and _plate.has_shrunken(plate):
                return True
        return False

    def undo_unique(self, unique_tensor: storch.Tensor) -> torch.Tensor:
        """
        Convert the unique tensor back to the non-unique format, then add the old plates back in
        # TODO: Make sure self.shrunken_plates is added
        # TODO: What if unique_tensor contains new plates after the unique?
        :param unique_tensor:
        :return:
        """
        plate_idx = unique_tensor.get_plate_dim_index(self.name)
        with storch.ignore_wrapping():
            dim_swapped = unique_tensor.transpose(
                plate_idx, unique_tensor.plate_dims - 1
            )
            fl_selected = torch.index_select(
                dim_swapped, dim=0, index=self.inv_indexing
            )
            selected = fl_selected.reshape(
                tuple(map(lambda p: p.n, self.shrunken_plates)) + fl_selected.shape[1:]
            )
            return storch.Tensor(
                selected,
                [unique_tensor],
                self.shrunken_plates + unique_tensor.plates,
                "undo_unique_" + unique_tensor.name,
            )


def unique(tensor: storch.Tensor, event_dim: Optional[int] = 0) -> storch.Tensor:
    with storch.ignore_wrapping():
        fl_tensor = torch.flatten(tensor, tensor.plate_dims)
        uniq, inverse_indexing = torch.unique(
            fl_tensor, return_inverse=True, dim=event_dim
        )
    inverse_indexing = storch.Tensor(
        inverse_indexing, [tensor], tensor.plates, "inv_index_" + tensor.name
    )
    uq_plate = UniquePlate(
        "uq_plate_" + tensor.name,
        uniq.shape[0],
        tensor.multi_dim_plates(),
        inverse_indexing,
    )
    return storch.Tensor(uniq, [tensor], [uq_plate], "unique_" + tensor.name)


def undo_unique(tensor: storch.Tensor) -> storch.Tensor:
    while is_unique(tensor):
        for plate in tensor.plates:
            if isinstance(plate, UniquePlate):
                tensor = plate.undo_unique(tensor)
    return tensor


def is_unique(tensor: storch.Tensor) -> bool:
    for plate in tensor.plates:
        if isinstance(plate, UniquePlate):
            return True
    return False



from typing import Optional, List, Union, Tuple

from storch.tensor import Tensor, StochasticTensor, CostTensor, IndependentTensor
import torch
from storch.util import print_graph, magic_box
import storch


_cost_tensors: [CostTensor] = []
_sampling_methods: [storch.method.Method] = []


def denote_independent(
    tensor: torch.Tensor,
    dim: int,
    plate_name: str,
    weight: Optional[storch.Tensor] = None,
) -> IndependentTensor:
    """
    Denote the given dimensions on the tensor as being independent, that is, batched dimensions.
    It will automatically put these dimensions to the left.
    :param tensor:
    :param dims:
    :param plate_name: Name of the plate. Reused if called again
    :return:
    """
    if (
        storch.wrappers._context_stochastic
        or storch.wrappers._context_deterministic > 0
    ):
        raise RuntimeError(
            "Cannot create independent tensors within a deterministic or stochastic context."
        )
    if isinstance(tensor, storch.Tensor):
        t_tensor = tensor._tensor
        if dim != 0:
            t_tensor = t_tensor.transpose(dim, 0)
        tensor_name = (
            tensor.name + "_indep_" + plate_name if tensor.name else plate_name
        )
        return IndependentTensor(
            t_tensor, [tensor], tensor.plates, tensor_name, plate_name, weight
        )
    else:
        if dim != 0:
            tensor = tensor.transpose(dim, 0)
        return IndependentTensor(tensor, [], [], plate_name, plate_name, weight)


def gather_samples(
    samples: Union[List[storch.Tensor], List[torch.Tensor]],
    plate_name: str,
    weight: Optional[storch.Tensor] = None,
) -> IndependentTensor:
    if (
        storch.wrappers._context_stochastic
        or storch.wrappers._context_deterministic > 0
    ):
        raise RuntimeError(
            "Cannot create independent tensors within a deterministic or stochastic context."
        )
    collect_tensors = []
    for sample in samples:
        if isinstance(sample, storch.Tensor):
            sample = sample._tensor
        sample = sample.unsqueeze(0)
        collect_tensors.append(sample)
    cat_tensors = storch.cat(collect_tensors, 0)

    if isinstance(samples[0], storch.Tensor):
        tensor_name = (
            samples[0].name + "_indep_" + plate_name if samples[0].name else plate_name
        )
        return IndependentTensor(
            cat_tensors,
            samples,
            samples[0].plates.copy(),
            tensor_name,
            plate_name,
            weight,
        )
    return IndependentTensor(
        cat_tensors, [], [], "_indep_" + plate_name, plate_name, weight
    )


def add_cost(cost: Tensor, name: str):
    if cost.event_shape != ():
        if cost.event_shape == (1,):
            cost = cost.squeeze(-1)
        else:
            raise ValueError("Can only register cost functions with empty event shapes")
    if not name:
        raise ValueError(
            "No name provided to register cost node. Make sure to register an unique name with the cost."
        )
    cost = CostTensor(cost._tensor, [cost], cost.plates, name)
    if torch.is_grad_enabled():
        storch.inference._cost_tensors.append(cost)
    return cost


def surrogate_loss(debug: bool=False) -> storch.Tensor:
    costs: [storch.Tensor] = storch.inference._cost_tensors
    if not costs:
        raise RuntimeError("No cost nodes registered for backward call.")
    if debug:
        print_graph(costs)

    # Sum of averages of cost node tensors
    surrogate_loss = 0.0

    # Loop over different cost nodes
    for c in costs:
        # Do not detach the weights when reducing. This is used in for example expectations to weight the
        # different costs.
        # reduced_cost = storch.reduce_plates(c, detach_weights=False)
        #
        # if print_costs:
        #     print(c.name, ":", reduced_cost._tensor.item())
        # total_cost += reduced_cost
        # Compute gradients for the cost nodes themselves, if they require one.
        # if reduced_cost.requires_grad:
        #     accum_loss += reduced_cost

        L = c._tensor.new_tensor(0.0)
        cost_loss = 0.0
        # Walk topologically through the graph
        # This is a parallelized implementation of Algorithm 1 in the paper
        for parent in c.walk_parents(depth_first=False, reverse=True):
            # Instance check here instead of parent.stochastic, as backward methods are only used on these.
            if not isinstance(parent, StochasticTensor):
                continue
            if not parent.requires_grad or not parent.method:
                continue

            if parent.method.is_pathwise(parent, c):
                continue
            # Transpose the parent stochastic tensor, so that its shape is the same as the cost but the event shape, and
            # possibly extra dimensions...?
            parent_tensor = parent._tensor
            reduced_cost = c
            parent_plates = parent.multi_dim_plates()
            # Reduce all plates that are in the cost node but not in the parent node
            for plate in storch.order_plates(c.multi_dim_plates(), reverse=True):
                if plate not in parent_plates:
                    reduced_cost = plate.reduce(reduced_cost, detach_weights=True)
            # Align the parent tensor so that the plate dimensions are in the same order as the cost tensor
            # TODO: This can probably be implemented with torch.movedim
            for index_c, plate in enumerate(reduced_cost.multi_dim_plates()):
                index_p = parent_plates.index(plate)
                if index_c != index_p:
                    parent_tensor = parent_tensor.transpose(index_p, index_c)
                    parent_plates[index_p], parent_plates[index_c] = (
                        parent_plates[index_c],
                        parent_plates[index_p],
                    )
            # Add empty (k=1) plates to new parent
            for plate in parent.plates:
                if plate not in parent_plates:
                    parent_plates.append(plate)

            # Create new storch Tensors with different order of plates for the cost and parent
            new_parent = storch.tensor.StochasticTensor(
                parent_tensor,
                [],
                parent_plates,
                parent.name,
                parent.n,
                parent.distribution,
                parent._requires_grad,
                parent.method,
            )
            new_parent.param_grads = parent.param_grads
            # Fake the new parent to be the old parent within the graph by mimicking its place in the graph
            new_parent._parents = parent._parents
            for p, has_link in new_parent._parents:
                p._children.append((new_parent, has_link))
            new_parent._children = parent._children

            # Compute the estimator
            (
                gradient_function,
                control_variate,
            ) = parent.method._estimator(new_parent, reduced_cost)

            if gradient_function is not None:
                L = L + gradient_function
            # Compute control variate
            if control_variate is not None:
                final_A = magic_box(L) * control_variate
                final_A = storch.reduce_plates(
                    final_A, detach_weights=False,  # TODO: Should this boolean be false or true?
                )
                if final_A.ndim == 1:
                    final_A = final_A.squeeze(0)
                cost_loss += final_A
        # Use magic box to distribute the cost
        cost_loss += storch.reduce_plates(magic_box(L) * c, detach_weights=False)
        # Sum over different costs
        surrogate_loss += cost_loss
    return surrogate_loss

def backward(
    debug: bool = False,
    create_graph: bool = False,
    update_estimator_params: bool = True,
) -> torch.Tensor:
    """
    Computes the gradients of the cost nodes with respect to the parameter nodes. It uses the storch
    methods used to sample stochastic nodes to properly estimate their gradient.

    Args:
        debug: Prints debug information on the backwards call.
        accum_grads: Saves gradient information in stochastic nodes. Note that this is an expensive option as it
        requires doing O(n) backward calls for each stochastic node sampled multiple times. Especially if this is a
        hierarchy of multiple samples.
        create_graph (bool): Creates the backpropagation graph of the gradient estimation for higher-order derivatives
    Returns:
        torch.Tensor: The average total cost normalized by the sampling weights.
    """

    SL = surrogate_loss(debug)
    _create_graph = create_graph

    stochastic_nodes = set()
    costs: [storch.Tensor] = storch.inference._cost_tensors
    for c in costs:
        for parent in c.walk_parents():
            # Instance check here instead of parent.stochastic, as backward methods are only used on these.
            if isinstance(parent, StochasticTensor):
                stochastic_nodes.add(parent)
                if parent.requires_grad and parent.method:
                    create_higher_order_graph = parent.method.should_create_higher_order_graph()
                    _create_graph = (
                            create_higher_order_graph or _create_graph
                    )
    if isinstance(SL, storch.Tensor) and SL._tensor.requires_grad:
        SL._tensor.backward(create_graph=_create_graph)

    if update_estimator_params:
        for s_node in stochastic_nodes:
            if s_node.method:
                s_node.method._update_parameters()

    if not create_graph:
        SL._clean()
        SL.grad_fn = None
        reset()

    return SL._tensor


def reset() -> None:
    # Free the SC graph links. This often improves garbage collection for larger graphs.
    # Unfortunately Python's GC seems to have imperfect cycle detection?
    for c in storch.inference._cost_tensors:
        c._clean()

    storch.inference._cost_tensors = []
    for method in storch.inference._sampling_methods:
        method.reset()
    storch.inference._sampling_methods = []



import torch
import torch.nn as nn
import storch
from torch.distributions import OneHotCategorical

torch.manual_seed(0)


class DiscreteVAE(nn.Module):
    def __init__(self):
        super().__init__()
        self.fc1 = nn.Linear(784, 512)
        self.fc2 = nn.Linear(512, 256)
        self.fc3 = nn.Linear(256, 2 * 10)
        self.fc4 = nn.Linear(2 * 10, 256)
        self.fc5 = nn.Linear(256, 512)
        self.fc6 = nn.Linear(512, 784)

    def encode(self, x):
        h1 = self.fc1(x).relu()
        h2 = self.fc2(h1).relu()
        return self.fc3(h2)

    def decode(self, z):
        h3 = self.fc4(z).relu()
        h4 = self.fc5(h3).relu()
        return self.fc6(h4).sigmoid()


def generative_story(
    method: storch.method.Method, model: DiscreteVAE, data: torch.Tensor
):
    x = storch.denote_independent(data.view(-1, 784), 0, "data")

    # Encode data. Shape: (data, 2 * 10)
    q_logits = model.encode(x)
    # Shape: (data, 2, 10)
    q_logits = q_logits.reshape(-1, 2, 10)
    q = OneHotCategorical(probs=q_logits.softmax(dim=-1))
    # Sample from variational posterior
    z = method(q)

    prior = OneHotCategorical(probs=torch.ones_like(q.probs) / 10.0)
    # Shape: (data)
    KL_div = torch.distributions.kl_divergence(q, prior).sum(-1)
    storch.add_cost(KL_div, "kl-div")

    z_in = z.reshape(z.shape[:-2] + (2 * 10,))
    reconstruction = model.decode(z_in)
    bce = torch.nn.BCELoss(reduction="none")(reconstruction, x).sum(-1)
    # bce = torch.nn.BCELoss(reduction="sum")(reconstruction, x)
    storch.add_cost(bce, "reconstruction")
    return z


from torchvision import datasets, transforms

train_loader = torch.utils.data.DataLoader(
    datasets.MNIST(
        "./data", train=True, download=True, transform=transforms.ToTensor(),
    ),
    shuffle=True,
    batch_size=64,
)


def train(method: storch.method.Method, train_loader):
    model = DiscreteVAE()
    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    for epoch in range(5):
        print("Epoch:" + str(epoch + 1))
        for i, (data, _) in enumerate(train_loader):
            # if i % 300 == 0:
            # evaluate(method, model, data, optimizer)
            optimizer.zero_grad()

            generative_story(method, model, data)
            elbo = storch.backward()
            optimizer.step()
            if i % 300 == 0:
                print("Training ELBO " + str(elbo.item()))


def evaluate(method: storch.method.Method, model: DiscreteVAE, data, optimizer):
    # Compute expected gradient
    optimizer.zero_grad()

    z = generative_story(storch.method.Expect("z"), model, data)
    storch.backward()
    expected_gradient = z.param_grads["probs"]

    # Collect gradient samples
    gradients = []
    for i in range(100):
        optimizer.zero_grad()

        z = generative_story(method, model, data)
        elbo = storch.backward()
        gradients.append(z.param_grads["probs"])

    gradients = storch.gather_samples(gradients, "gradients")
    mean_gradient = storch.reduce_plates(gradients, "gradients")
    bias_gradient = (
        storch.reduce_plates((mean_gradient - expected_gradient) ** 2)
    ).sum()
    print(
        "Training ELBO "
        + str(elbo.item())
        + " Gradient variance "
        + str(storch.variance(gradients, "gradients")._tensor.item())
        + " Gradient bias "
        + str(bias_gradient._tensor.item())
    )


#
# train(
#     storch.method.ScoreFunction("z", n_samples=10, baseline_factory="batch_average"),
#     train_loader,
# )


train(
    storch.method.Expect("z"), train_loader,
)



from typing import Optional, Tuple

import storch
from storch.method.method import Method
from storch.sampling import UnorderedSet, GumbelSoftmaxWOR
from storch.util import magic_box


class UnorderedSetEstimator(Method):
    """
    Implements the Unordered Set REINFORCE Estimator with baseline from https://openreview.net/forum?id=rklEj2EFvB
    (Wouter Kool et al, ICLR 2020)
    It uses Stochastic Beam Search to sample from sequences of discrete random variables without replacement.
    Then it normalizes these samples, treating them as unordered samples without replacement, that is, as sets.
    The baseline takes a weighted average over the cost using these samples.

    This implementation uses the trapezoid rule to compute the leave-one-out ratio for the plate weighting.
    This numerical integration can be demanding. One can choose to decrease num_int_points to trade off computation
    time for precision.
    """

    def __init__(
        self,
        plate_name: str,
        k: int,
        use_baseline: bool = True,
        exact_integration: bool = False,
        num_int_points: int = 1000,
        a: float = 5.0,
        eos=None,
    ):
        """
        Creates an Unordered Set Estimator method.
        :param plate_name: The name of the ancestral plate for the sequence to be sampled.
        :param k: The amount of samples to take using stochastic beam search.
        :param use_baseline: Whether to use the built-in baseline (see Equation 17 and 18 of https://openreview.net/forum?id=rklEj2EFvB)
        :param exact_integration: Whether to use exact integration instead of numerical integration. Currently not implemented
        :param num_int_points: How many points to use in the trapezoid rule for the numerical integration in the
        computation of the leave-one-out ratio.
        :param a: Hyperparameter for numerical stable computation of the leave-one-out ratio. The default is recommended
        in https://openreview.net/forum?id=rklEj2EFvB, but could be tuned if NaNs pop up.
        """
        super().__init__(
            plate_name,
            UnorderedSet(
                plate_name,
                k,
                use_baseline,
                exact_integration,
                num_int_points,
                a,
                eos=eos,
            ),
        )
        self.use_baseline = use_baseline

    def is_pathwise(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> bool:
        # We only want to add a loss on the stochastic tensor with the same plate as the cost node.
        # This is because the estimator computes the gradient with the respect to the JOINT log probability.
        # If we would have added the gradient for all stochastic tensors, these would just be duplicates of the same
        # loss being added (ie that gradient would be oversampled)
        for distr_plate in tensor.plates:
            if distr_plate.name == self.plate_name:
                for cost_plate in cost_node.plates:
                    if cost_plate.name == self.plate_name:
                        if cost_plate is distr_plate:
                            return False
                        return True
                raise ValueError(
                    "The given tensor contains an ancestral plate that the cost node doesn't have."
                )
        return True

    def estimator(
        self, tensor: storch.StochasticTensor, cost_node: storch.CostTensor
    ) -> Tuple[
        Optional[storch.Tensor], Optional[storch.Tensor]
    ]:
        # Note: We automatically multiply with leave-one-out ratio in the plate reduction
        plate = None
        for _p in cost_node.plates:
            if _p.name == self.plate_name:
                plate = _p
                break
        if not self.use_baseline:
            return plate.log_probs * cost_node.detach()

        # Subtract the 'average' cost of other samples, keeping in mind that samples are not independent.
        # plates x k
        baseline = storch.sum(
            (plate.log_probs + plate.log_snd_leave_one_out).exp() * cost_node, plate
        )
        # Make sure the k dimension is recognized as batch dimension
        baseline = storch.Tensor(
            baseline._tensor, [baseline], baseline.plates + [plate]
        )
        return plate.log_probs, (1 - magic_box(plate.log_probs)) * baseline.detach()


class UnorderedSetGumbelSoftmax(Method):
    def __init__(
        self,
        plate_name: str,
        k: int,
        initial_temperature=1.0,
        min_temperature=1.0e-4,
        annealing_rate=1.0e-5,
        eos=None,
    ):
        super().__init__(
            plate_name,
            GumbelSoftmaxWOR(
                plate_name, k, initial_temperature, min_temperature, annealing_rate, eos
            ),
        )



from __future__ import annotations
import torch
import storch
from torch.distributions import Distribution
from collections import deque
from typing import List, Iterable, Any, Callable, Iterator, Dict
import builtins
from itertools import product
from typing import Optional
from storch.exceptions import IllegalStorchExposeError
from storch.excluded_init import (
    exception_methods,
    excluded_methods,
    unwrap_only_methods,
    expand_methods,
)

# from storch.typing import BatchTensor


class Plate:
    def __init__(
        self,
        name: str,
        n: int,
        parents: List[Plate],
        weight: Optional[storch.Tensor] = None,
    ):
        self.weight = weight
        if weight is None:
            self.weight = torch.tensor(1.0 / n)
        self.name = name
        self.n = n
        self.parents = parents

    def __eq__(self, other) -> bool:
        if not isinstance(other, Plate):
            return False
        if self.name != other.name:
            return False
        if self.n != other.n:
            # TODO: This should maybe return an error...?
            return False
        if isinstance(self.weight, storch.Tensor):
            if not isinstance(other.weight, storch.Tensor):
                return False
            if self.weight._tensor is other.weight._tensor:
                return True
            if self.weight.shape != other.weight.shape:
                return False
            return self.weight._tensor.__eq__(other.weight._tensor).all()
        if isinstance(other.weight, storch.Tensor):
            return False
        # Neither of the weights are Tensors, so the weights must be equal as self.n==other.n
        return True

    def __str__(self):
        return self.name + ", " + str(self.n)

    def __repr__(self):
        return (
            "("
            + self.name.__repr__()
            + ", "
            + self.n.__repr__()
            + ", "
            + self.weight.__repr__()
            + ")"
        )

    def reduce(self, tensor: storch.Tensor, detach_weights=True):
        plate_weighting = self.weight
        if detach_weights:
            plate_weighting = self.weight.detach()
        # Case: The weight is a single number. First sum, then multiply with the weight (usually taking the mean)
        if plate_weighting.ndim == 0:
            return storch.sum(tensor, self) * plate_weighting

        # Case: There is a weight for each plate which is not dependent on the other batch dimensions
        elif plate_weighting.ndim == 1:
            index = tensor.get_plate_dim_index(self.name)
            plate_weighting = plate_weighting[
                (...,) + (None,) * (tensor.ndim - index - 1)
            ]
            weighted_tensor = tensor * plate_weighting
            return storch.sum(weighted_tensor, self)

        # Case: The weight is a vector of numbers equal to batch dimension. Assumes it is a storch.Tensor
        else:
            for parent_plate in self.parents:
                if parent_plate not in tensor.plates:
                    raise ValueError(
                        "Plate missing when reducing tensor: " + parent_plate.name
                    )
            weighted_tensor = tensor * plate_weighting
            return storch.sum(weighted_tensor, self)

    def on_collecting_args(self, plates: [Plate]) -> bool:
        """
        Gets called after a wrapper collected plates from its input arguments.

        Args:
            plates ([Plate]):  All collected plates

        Returns:
            bool: True if this plate should remain in the collected plates.

        """
        for plate in plates:
            if plate.name == self.name and plate != self:
                if not plate.on_duplicate_plate(self):
                    return False
        return True

    def on_duplicate_plate(self, plate: Plate) -> bool:
        raise ValueError("")

    def on_unwrap_tensor(self, tensor: storch.Tensor) -> storch.Tensor:
        """
        Gets called whenever the given tensor is being unwrapped and unsqueezed for batch use.

        Args:
            tensor (storch.Tensor): The input tensor that is being unwrapped

        Returns:
            storch.Tensor: The tensor that will be unwrapped and unsqueezed in the future. Can be a modification of the input tensor.

        """
        return tensor


class Tensor:
    """
    A :class:`storch.Tensor` is a wrapper around a :class:`torch.Tensor` that acts like a normal :class:`torch.Tensor`
    with some restrictions and some extra data.

    By design, :class:`storch.Tensor` cannot expose the wrapped :class:`torch.Tensor` to regular Python control flow.
    For example, using a :class:`storch.Tensor` inside an ``if`` condition or in a ``for`` loop will throw an
    :class:`~storch.exceptions.IllegalStorchExposeError`. This is done because a node could be dependent on a Tensor that is used as
    a conditional to branch between different computation paths.  However, Python control flow will not register dependencies
    between nodes in the computation graph.

    The underlying :class:`torch.Tensor` can be unwrapped in two ways. The safe way is using the :func:`.deterministic`
    wrapper, which safely unwraps the :class:`storch.Tensor` and runs the function on the unwrapped :class:`torch.Tensor`.
    Note that all ``torch`` methods are automatically wrapped using :func:`.deterministic` when an input argument
    is :class:`storch.Tensor`.
    The unsafe way to unwrap the tensor is to access :attr:`storch.Tensor._tensor`. This should only be use when one is
    sure this will not introduce missing dependency links.

    Args:
        tensor (torch.Tensor): The tensor to wrap. The leftmost dimensions should correspond to the sizes of ``plates``
            that are larger than 1.
        parents ([storch.Tensor]): The parents of this Tensor. Parents represent the incoming links in stochastic
            computation graphs.
        plates ([storch.Plate]): The plates of this Tensor. Plates contain information about the sampling procedure and
            dependencies of this Tensor with respect to earlier samples.
        name (Optional[str]): The name of this Tensor.
    """

    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        name: Optional[str] = None,
    ):
        if isinstance(tensor, Tensor):
            raise TypeError(
                "storch.Tensors should be constructed with torch.Tensors, not other storch.Tensors."
            )
        plate_names = set()
        batch_dims = 0
        # Check whether this tensor does not violate the constraints imposed by the given batch_links
        for plate in plates:
            if plate.name in plate_names:
                raise ValueError(
                    "Plates contain two instances of same plate "
                    + plate.name
                    + ". This can be caused by different samples with the same name using a different amount of "
                    + "samples n or different weighting of the samples. Make sure that these samples use the same number of samples."
                )
            plate_names.add(plate.name)
            # plate length is 1. Ignore this dimension, as singleton dimensions should not exist.
            if plate.n == 1:
                continue
            if len(tensor.shape) <= batch_dims:
                raise ValueError(
                    "Got an input tensor with too few dimensions. We expected "
                    + str(len(plates))
                    + " plate dimensions. Instead, we found only "
                    + str(len(tensor.shape))
                    + " dimensions. Violated at dimension "
                    + str(batch_dims)
                )
            elif not tensor.shape[batch_dims] == plate.n:
                raise ValueError(
                    "Storch Tensors should take into account their surrounding plates. Violated at dimension "
                    + str(batch_dims)
                    + " and plate "
                    + plate.name
                    + " with size "
                    + str(plate.n)
                    + ". "
                    "Instead, it was "
                    + str(tensor.shape[batch_dims])
                    + ". Batch links: "
                    + str(plates)
                    + " Tensor shape: "
                    + str(tensor.shape)
                )
            batch_dims += 1

        self._name = name
        self._tensor = tensor
        self._parents = []
        self._cleaned = False
        for p in parents:
            # TODO: Should I re-add this?
            # if p.is_cost:
            #     raise ValueError("Cost nodes cannot have children.")
            differentiable_link = has_backwards_path(self, p)
            self._parents.append((p, differentiable_link))
            p._children.append((self, differentiable_link))
        self._children = []
        self.plate_dims = batch_dims
        self.event_shape = tensor.shape[batch_dims:]
        self.event_dims = len(self.event_shape)
        self.plates = plates

    def __torch_function__(self, func, types, args=(), kwargs=None):
        """
        Called whenever a torch.* or torch.nn.functional.* method is being called on a storch.Tensor. This wraps
        that method in the deterministic wrapper to properly handle all input arguments and outputs.
        """
        if kwargs is None:
            kwargs = {}
        func_name = func.__name__
        if func_name in exception_methods:
            raise IllegalStorchExposeError(
                "Calling method " + func_name + " with storch tensors is not allowed."
            )
        if func_name in excluded_methods:
            return func(*args, **kwargs)

        if func_name in expand_methods:
            # Automatically expand empty plate dimensions. This is necessary for some loss functions, which
            # assume both inputs have exactly the same elements.
            return storch.wrappers._handle_deterministic(
                func, args, kwargs, expand_plates=True
            )
        # if func_name in unwrap_only_methods:
        #     return storch.wrappers._unpack_wrapper(func)(*args, *kwargs)

        return storch.wrappers._handle_deterministic(func, args, kwargs)

    def __getattr__(self, item):
        """
        Called whenever an attribute is called on a storch.Tensor object that is not directly implemented by storch.Tensor.
        It defers it to the underlying torch.Tensor. If it is a callable (ie, torch.Tensor implements a function
        with the name item), it will wrap this callable with a deterministic wrapper.

        TODO: This should probably filter the methods
        """
        attr = getattr(torch.Tensor, item)
        if isinstance(attr, Callable):
            func_name = attr.__name__
            if func_name in exception_methods:
                raise IllegalStorchExposeError(
                    "Calling method "
                    + func_name
                    + " with storch tensors is not allowed."
                )
            if func_name in excluded_methods:
                return attr
            # if func_name in unwrap_only_methods:
            #     return storch.wrappers._unpack_wrapper(attr, self=self)
            return storch.wrappers._self_deterministic(attr, self)

    @property
    def name(self) -> str:
        return self._name

    @property
    def is_sparse(self) -> bool:
        """
        Returns: True if the underlying tensor is sparse.
        """
        return self._tensor.is_sparse

    def __str__(self) -> str:
        t = (
            (self.name + ": " if self.name else "") + "Stochastic"
            if self.stochastic
            else ("Cost" if self.is_cost else "Deterministic")
        )
        return t + " " + str(self._tensor) + " Batch links: " + str(self.plates)

    def __repr__(self):
        return self._tensor.__repr__()

    def __hash__(self):
        return object.__hash__(self)

    @storch.deterministic
    def __eq__(self, other):
        return self.__eq__(other)

    @storch.deterministic(l_broadcast=False)
    def __getitem__(self, index):
        return self.__getitem__(index)

    @storch.deterministic(l_broadcast=False)
    def __setitem__(self, index, value):
        return self.__setitem__(index, value)

    def _walk_backwards(
        self,
        expand_fn,
        depth_first=True,
        reverse=False, # Only supported for breadth-first
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ) -> Iterator:
        visited = set()
        visited_ordered = []
        if depth_first:
            S = [self]
            while S:
                v = S.pop()
                if repeat_visited or v not in visited:
                    yield walk_fn(v)
                    visited.add(v)
                    for w, d in expand_fn(v):
                        if d or not only_differentiable:
                            S.append(w)
        else:
            queue = deque()
            visited.add(self)
            queue.append(self)
            while queue:
                v = queue.popleft()
                if reverse:
                    visited_ordered.append(v)
                else:
                    yield walk_fn(v)
                for w, d in expand_fn(v):
                    if (repeat_visited or w not in visited) and (
                        d or not only_differentiable
                    ):
                        visited.add(w)
                        queue.append(w)
            if reverse:
                return reversed(visited_ordered)

    def walk_parents(
        self,
        depth_first=True,
        reverse=False,
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ):
        """
        Searches through the parents of this Tensor in the stochastic computation graph.

        Args:
            depth_first: True to use depth first, otherwise breadth first.
            reverse: Reverse the order: If true, instead of first returning the immediate parents, return
                the parents furthest up, working towards the immediate parents.
                Currently only supported for breadth-first search.
            only_differentiable: True to only walk over edges that are differentiable
            repeat_visited:
            walk_fn: Optional function on :class:`storch.Tensor` that manipulates the nodes found.

        Returns:
            Iterator of type that is equal to the output type of ``walk_fn``.
        """
        return self._walk_backwards(
            lambda p: p._parents,
            depth_first,
            only_differentiable,
            repeat_visited,
            walk_fn,
        )

    def walk_children(
        self,
        depth_first=True,
        only_differentiable=False,
        repeat_visited=False,
        walk_fn=lambda x: x,
    ):
        """
        Searches through the children of this Tensor in the stochastic computation graph.

        Args:
            depth_first: True to use depth first, otherwise breadth first.
            only_differentiable: True to only walk over edges that are differentiable
            repeat_visited:
            walk_fn: Optional function on :class:`storch.Tensor` that manipulates the nodes found.

        Returns:
            Iterator of type that is equal to the output type of ``walk_fn``.
        """
        return self._walk_backwards(
            lambda p: p._children,
            depth_first,
            only_differentiable,
            repeat_visited,
            walk_fn,
        )

    def _clean(self):
        """
        Cleans up :attr:`_children` and :attr:`_parents` for all nodes in the subgraph of this node (depth first)
        """
        if self._cleaned:
            return
        self._cleaned = True
        for (node, _) in self._children:
            node._clean()
        for (node, _) in self._parents:
            node._clean()
        self._children = []
        self._parents = []

    def detach_tensor(self) -> storch.Tensor:
        """
        Returns: A :class:`storch.Tensor` that is removed from PyTorch's differention graph.
            However, the tensor will remain present on the stochastic computation graph.
        """
        return self._tensor.detach()

    @property
    def stochastic(self) -> bool:
        """
        Returns:
            bool: True if this is a stochastic node in the stochastic computation graph, False otherwise.
        """
        return False

    @property
    def is_cost(self) -> bool:
        """
        Returns:
            bool: True if this is a cost node in the stochastic computation graph, False otherwise.
        """
        return False

    @property
    def requires_grad(self) -> bool:
        return self._tensor.requires_grad

    @property
    def plate_shape(self) -> torch.Size:
        return self._tensor.shape[: self.plate_dims]

    def size(self, *args) -> torch.Size:
        return self._tensor.size(*args)

    @property
    def shape(self) -> torch.Size:
        return self._tensor.size()

    def is_cuda(self):
        return self._tensor.is_cuda

    @property
    def dtype(self):
        return self._tensor.dtype

    @property
    def layout(self):
        return self._tensor.layout

    @property
    def device(self):
        return self._tensor.device

    @property
    def grad(self):
        return self._tensor.grad

    def dim(self):
        return self._tensor.dim()

    def ndimension(self):
        return self._tensor.ndimension()

    @property
    def ndim(self):
        return self._tensor.ndim

    def register_hook(self, hook: Callable) -> Any:
        return self._tensor.register_hook(hook)

    @property
    def event_dim_indices(self):
        return range(self.plate_dims, self._tensor.dim())

    def get_plate(self, plate_name: str) -> Plate:
        for plate in self.plates:
            if plate.name == plate_name:
                return plate
        raise IndexError("Tensor has no such plate: " + plate_name + ".")

    def get_plate_dim_index(self, plate_name: str) -> int:
        for i, plate in enumerate(self.multi_dim_plates()):
            if plate.name == plate_name:
                return i
        raise IndexError(
            "Tensor has no such plate: "
            + plate_name
            + ". Alternatively, the dimension of this batch is 1."
        )

    def iterate_plate_indices(self) -> Iterable[List[int]]:
        ranges = list(map(lambda a: list(range(a)), self.plate_shape))
        return product(*ranges)

    def multi_dim_plates(self) -> List[Plate]:
        return list(filter(lambda p: p.n > 1, self.plates))

    def backward(
        self,
        gradient: Optional[Tensor] = None,
        keep_graph: bool = False,
        create_graph: bool = False,
        retain_graph: bool = False,
    ) -> None:
        raise NotImplementedError(
            "Cannot call .backward on storch.Tensor. Instead, register cost nodes using "
            "storch.add_cost, then use storch.backward()."
        )

    # region OperatorOverloads

    def __len__(self) -> int:
        return self._tensor.__len__()

    def __index__(self) -> int:
        raise IllegalStorchExposeError("Cannot use storch tensors as index.")

    @storch.deterministic
    def eq(self, other) -> bool:
        return self.eq(other)

    def __getstate__(self):
        raise NotImplementedError(
            "Pickle is currently not implemented for storch tensors."
        )

    def __setstate__(self, state):
        raise NotImplementedError(
            "Pickle is currently not implemented for storch tensors."
        )

    def __bool__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to boolean. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a boolean."
        )

    def __float__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to float. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a float."
        )

    def __int__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to int. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as an int."
        )

    def __long__(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to long. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a long."
        )

    def __nonzero__(self) -> builtins.bool:
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to boolean. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a boolean."
        )

    def __array__(self):
        self.numpy()

    def __array_wrap__(self):
        self.numpy()

    def numpy(self):
        raise IllegalStorchExposeError(
            "It is not allowed to convert storch tensors to numpy arrays. Make sure to unwrap "
            "storch tensors to normal torch tensor to use this tensor as a np.array."
        )

    def __contains__(self, item):
        raise IllegalStorchExposeError(
            "It is not allowed to expose storch tensors via in statements."
        )

    def __deepcopy__(self, memodict={}):
        raise NotImplementedError(
            "There is currently no deep copying implementation for storch Tensors."
        )

    def __iter__(self):
        # TODO: This recognizes storch.Tensor as Iterable, even though it's not implemented.
        raise NotImplementedError("Cannot currently iterate over storch Tensors.")

    def detach_(self) -> Tensor:
        raise NotImplementedError("In place detach is not allowed on storch tensors.")

    def __add__(self, other):
        return torch.add(self, other)

    def __radd__(self, other):
        return torch.add(other, self)

    def __sub__(self, other):
        return torch.sub(self, other)

    def __rsub__(self, other):
        return torch.sub(other, self)

    def __mul__(self, other):
        return torch.mul(self, other)

    def __rmul__(self, other):
        return torch.mul(other, self)

    def __matmul__(self, other):
        return torch.matmul(self, other)

    def __rmatmul__(self, other):
        return torch.matmul(other, self)

    def __pow__(self, other):
        return torch.pow(self, other)

    def __rpow__(self, other):
        return torch.pow(other, self)

    def __div__(self, other):
        return torch.div(self, other)

    def __rdiv__(self, other):
        return torch.div(other, self)

    def __mod__(self, other):
        return torch.remainder(self, other)

    def __rmod__(self, other):
        return torch.remainder(other, self)

    def __truediv__(self, other):
        return torch.true_divide(self, other)

    def __rtruediv__(self, other):
        return torch.true_divide(other, self)

    def __floordiv__(self, other):
        return torch.floor_divide(self, other)

    def __rfloordiv__(self, other):
        return torch.floor_divide(other, self)

    def __abs__(self):
        return torch.abs(self)

    def __and__(self, other):
        return torch.logical_and(self, other)

    def __rand__(self, other):
        return torch.logical_and(other, self)

    def __ge__(self, other):
        return torch.ge(self, other)

    def __gt__(self, other):
        return torch.gt(self, other)

    @storch.deterministic
    def __invert__(self):
        return self.__invert__()

    def __le__(self, other):
        return torch.le(self, other)

    @storch.deterministic
    def __lshift__(self, other):
        return self.__lshift__(other)

    @storch.deterministic
    def __lshift__(self, other):
        return other.__lshift__(self)

    def __lt__(self, other):
        return torch.lt(self, other)

    def ne(self, other):
        return torch.ne(self, other)

    def __neg__(self):
        return torch.neg(self)

    def __or__(self, other):
        return torch.logical_or(self, other)

    def __ror__(self, other):
        return torch.logical_or(other, self)

    def __pos__(self):
        # TODO: Is this correct?
        return self

    @storch.deterministic
    def __rshift__(self, other):
        return self.__rshift__(other)

    @storch.deterministic
    def __rrshift__(self, other):
        return other.__rshift__(self)

    def __xor__(self, other):
        return torch.logical_xor(self, other)

    def __rxor__(self, other):
        return torch.logical_xor(other, self)

    # endregion


class CostTensor(Tensor):
    def __init__(self, tensor: torch.Tensor, parents, plate_links: [Plate], name: str):
        super().__init__(tensor, parents, plate_links, name)

    @property
    def is_cost(self) -> bool:
        return True


class IndependentTensor(Tensor):
    """
    Used to denote independencies on a Tensor. This could for example be the minibatch dimension. The first dimension
    of the input tensor is taken to be independent and added as a batch dimension to the storch system.
    """

    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        tensor_name: str,
        plate_name: str,
        weight: Optional[storch.Tensor],
    ):
        n = tensor.shape[0]
        for plate in plates:
            if plate.name == plate_name:
                raise ValueError(
                    "Cannot create independent tensor with name "
                    + plate_name
                    + ". A parent sample has already used"
                    " this name. Use a different name for this independent dimension."
                )
        plates.insert(0, Plate(plate_name, n, plates.copy(), weight))
        super().__init__(tensor, parents, plates, tensor_name)
        self.n = n

    # TODO: Should IndependentTensors be stochastic? Sometimes, like if it is denoting a minibatch, making them
    #  stochastic seems like it is correct. Are there other cases?
    def stochastic(self) -> bool:
        return True


class StochasticTensor(Tensor):
    """
    A :class:`storch.Tensor` that represents a stochastic node in the stochastic computation graph.

    Args:
        n (int): The size of the plate dimension created by this stochastic node.
        distribution: The distribution of this stochastic node.
        requires_grad (bool): True if we are interested in the gradients with respect to the parameters of the distribution of
            this stochastic node.
        
    """

    # TODO: Copy original tensor to make sure it cannot change using inplace
    def __init__(
        self,
        tensor: torch.Tensor,
        parents: [Tensor],
        plates: [Plate],
        name: str,
        n: int,
        distribution: Distribution,
        requires_grad: bool,
        method: Optional[storch.method.Method] = None,
    ):
        self.distribution = distribution
        super().__init__(tensor, parents, plates, name)
        self._requires_grad = requires_grad
        self.n = n
        self.method = method
        self.param_grads = {}
        self._grad = None
        self._clean_hooks = []

    @property
    def stochastic(self) -> bool:
        return True

    @property
    # TODO: Should not manually override it like this. The stochastic "requires_grad" should be a different method, so
    # that the meaning of requires_grad is consistent everywhere
    def requires_grad(self) -> bool:
        return self._requires_grad

    @property
    def grad(self) -> Dict[str, storch.Tensor]:
        return self.param_grads

    def _set_method(self, method: storch.method.Method):
        self.method = method

    def _clean(self) -> None:
        new_param_grads = {}
        for name, grad in self.param_grads.items():
            # In case higher-order derivatives are stored, remove these from the graph.
            new_param_grads[name] = grad.detach()
        for clean_hook in self._clean_hooks:
            clean_hook()
        self._clean_hooks = []
        super()._clean()


is_tensor = lambda a: isinstance(a, torch.Tensor) or isinstance(a, Tensor)
from storch.util import has_backwards_path



from __future__ import annotations
import torch
import storch
from typing import Union, List, Tuple, Callable
from torch import Size
from torch.distributions import Bernoulli, Categorical, OneHotCategorical

# from storch.seq import BlackboxTensor

AnyTensor = Union[storch.Tensor, torch.Tensor]

_size = Union[Size, List[int], Tuple[int, ...]]

Dims = Union[int, _size]

# AnyBlackboxTensor = Union[BlackboxTensor, torch.Tensor]

DiscreteDistribution = Union[Bernoulli, Categorical, OneHotCategorical]


_index = Union[str, int, storch.Plate]
_indices = Union[List[_index], _index]

_plates = Union[List[Union[storch.Plate, str]], storch.Plate, str]
